# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pennylane as qml
import jax
import jax.numpy as jnp
import numpy as np
import logging
from qcml.utils.jax import min_max_scale

logger = logging.getLogger(__name__)
jax.config.update("jax_enable_x64", True)


def separable_kernel(
    x1,
    x2,
    encoding_layers=1,
    jit=False,
    dev_type="default.qubit.jax",
    qnode_kwargs={"interface": "jax", "diff_method": None},
):
    n_qubits = x1.shape[0]

    def construct_circuit():
        # Directly define the observable without re-checking for Hermiticity
        projector = np.array([[1.0, 0.0], [0.0, 0.0]])
        dev = qml.device(dev_type, wires=1)

        @qml.qnode(dev, **qnode_kwargs)
        def qubit_circuit(x):
            for layer in range(encoding_layers):
                qml.RX(-np.pi / 4, wires=0)
                qml.RY(-x[0], wires=0)
            for layer in range(encoding_layers):
                qml.RY(x[1], wires=0)
                qml.RX(np.pi / 4, wires=0)
            return qml.expval(qml.Hermitian(projector, wires=0))

        return qubit_circuit

    # Initialize the circuit
    qubit_circuit = construct_circuit()

    def full_circuit(x):
        probs = [
            qubit_circuit(jnp.array([x[i], x[i + n_qubits]])) for i in range(n_qubits)
        ]
        return jnp.prod(jnp.array(probs))

    if jit:
        full_circuit = jax.jit(full_circuit)

    # Combine x1 and x2 to form the input to the circuit
    z = jnp.concatenate((x1, x2))

    # Compute the kernel value
    kernel_value = full_circuit(z)

    return kernel_value


def projected_quantum_kernel(
    x1,
    x2,
    embedding="Hamiltonian",
    t=1.0 / 3,
    trotter_steps=5,
    gamma_factor=1.0,
    jit=False,
    dev_type="default.qubit.jax",
    qnode_kwargs={"interface": "jax-jit", "diff_method": None},
):
    n_features = x1.shape[0]

    # Determine the number of qubits based on the embedding type
    if embedding == "IQP":
        n_qubits = n_features

        def embedding_fn(x):
            qml.IQPEmbedding(x, wires=range(n_qubits), n_repeats=2)

    elif embedding == "Hamiltonian":
        n_qubits = n_features + 1
        rotation_angles = jnp.array(np.random.uniform(size=(n_qubits, 3)) * np.pi * 2)
        evol_time = t / trotter_steps * (n_qubits - 1)

        def embedding_fn(x):
            for i in range(n_qubits):
                qml.Rot(
                    rotation_angles[i, 0],
                    rotation_angles[i, 1],
                    rotation_angles[i, 2],
                    wires=i,
                )
            for __ in range(trotter_steps):
                for j in range(n_qubits - 1):
                    qml.IsingXX(x[j] * evol_time, wires=[j, j + 1])
                    qml.IsingYY(x[j] * evol_time, wires=[j, j + 1])
                    qml.IsingZZ(x[j] * evol_time, wires=[j, j + 1])

    # Initialize the quantum device
    dev = qml.device(dev_type, wires=n_qubits)

    # Define the quantum circuit within a QNode
    @qml.qnode(dev, **qnode_kwargs)
    def circuit(x):
        embedding_fn(x)
        return (
            [qml.expval(qml.PauliX(wires=i)) for i in range(n_qubits)]
            + [qml.expval(qml.PauliY(wires=i)) for i in range(n_qubits)]
            + [qml.expval(qml.PauliZ(wires=i)) for i in range(n_qubits)]
        )

    # Optionally JIT compile the circuit
    if jit:
        circuit = jax.jit(circuit)

    # Compute expectation values for both input vectors
    expvals_x1 = jnp.array(circuit(x1))
    expvals_x2 = jnp.array(circuit(x2))

    # Compute the squared differences
    diff = expvals_x1 - expvals_x2
    diff_squared_sum = jnp.sum(diff**2)

    # Calculate default gamma value based on the variance of the expectation values
    default_gamma = 1 / jnp.var(expvals_x1) / n_features

    # Compute the final kernel value using the Gaussian kernel formula
    kernel_value = jnp.exp(-default_gamma * gamma_factor * diff_squared_sum)

    return kernel_value


def iqp_embedding_kernel(
    x1,
    x2,
    repeats=1,
    jit=False,
    dev_type="default.qubit.jax",
    qnode_kwargs={"interface": "jax-jit", "diff_method": None},
):
    n_qubits = x1.shape[0]  # Number of qubits equals the number of features
    dev = qml.device(dev_type, wires=n_qubits)
    # logger.debug(f"IQP embedding kernel is using {n_qubits} qubits")

    @qml.qnode(dev, **qnode_kwargs)
    def circuit(x1, x2):
        qml.IQPEmbedding(x1, wires=range(n_qubits), n_repeats=repeats)
        qml.adjoint(qml.IQPEmbedding(x2, wires=range(n_qubits), n_repeats=repeats))
        return qml.expval(
            qml.PauliZ(0)
        )  # Expectation value of the Pauli-Z operator on the first qubit

    if jit:
        circuit = jax.jit(circuit)

    kernel_value = circuit(x1, x2)

    # Add a small jitter to prevent identical values
    kernel_value += 1e-8 * jnp.sign(kernel_value)

    return kernel_value


def angle_embedding_kernel(
    x1,
    x2,
    rotation="Y",
    jit=False,
    dev_type="default.qubit.jax",
    qnode_kwargs={"interface": "jax-jit", "diff_method": None},
):
    n_qubits = x1.shape[0]
    dev = qml.device(dev_type, wires=n_qubits)
    # logger.debug(f"Angle embedding kernel is using {n_qubits} qubits")

    @qml.qnode(dev, **qnode_kwargs)
    def circuit(x1, x2):
        qml.AngleEmbedding(x1, wires=range(n_qubits), rotation=rotation)
        qml.adjoint(qml.AngleEmbedding)(x2, wires=range(n_qubits), rotation=rotation)
        return qml.expval(qml.PauliZ(0))

    if jit:
        circuit = jax.jit(circuit)

    return circuit(x1, x2)


def amplitude_embedding_kernel(
    x1,
    x2,
    pad_with=0.0,
    normalize=True,  # This controls the normalization within the embedding
    jit=False,
    dev_type="default.qubit.jax",
    qnode_kwargs={"interface": "jax-jit", "diff_method": None},
):
    n_qubits = int(np.ceil(np.log2(max(len(x1), len(x2)))))
    dev = qml.device(dev_type, wires=n_qubits)

    @qml.qnode(dev, **qnode_kwargs)
    def circuit(x1, x2, n_qubits):
        logger.info(f"Amplitude circuit has {n_qubits} qubits")
        qml.AmplitudeEmbedding(
            x1, wires=range(n_qubits), pad_with=pad_with, normalize=normalize
        )
        qml.adjoint(qml.AmplitudeEmbedding)(
            x2, wires=range(n_qubits), pad_with=pad_with, normalize=normalize
        )
        return qml.expval(qml.PauliZ(0))

    if jit:
        circuit = jax.jit(circuit)

    return circuit(x1, x2, n_qubits)



import os
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import json
import re
import numpy as np


class GridSearchAnalysis:
    def __init__(self, path):
        self.path = path
        self.df = None
        self._load_csv()

    def _clean_json(self, json_str):
        json_str = json_str.replace("'", '"')
        json_str = re.sub(r"(\w+):", r'"\1":', json_str)
        try:
            return json.loads(json_str)
        except json.JSONDecodeError:
            print(f"Failed to parse JSON: {json_str}")
            return {}

    def _load_csv(self):
        self.df = pd.read_csv(self.path)
        self.df["kernel_params"] = self.df["kernel_params"].apply(
            lambda x: self._clean_json(x) if pd.notnull(x) else {}
        )
        self.df["transf_params"] = self.df["transf_params"].apply(
            lambda x: self._clean_json(x) if pd.notnull(x) else {}
        )
        print(f"CSV file loaded from {self.path}")

    def plot_kernel_accuracy_range(
        self,
        experiment_name,
        classifier,
        groups,
        group_names=None,
        figsize=(5, 3),
        title=None,
        legend_title=None,
        palette="husl",
    ):
        if self.df is None:
            print("Dataframe is empty. Please load the CSV file first.")
            return

        df_filtered = self.df.loc[
            (self.df["experiment_name"] == experiment_name)
            & (self.df["classifier"] == classifier)
        ].copy()

        # Ensure that the 'kernel_func' column contains only string values
        df_filtered["kernel_func"] = df_filtered["kernel_func"].astype(str)

        # Now safely apply the .str methods
        df_filtered["kernel_func"] = (
            df_filtered["kernel_func"]
            .str.replace("_kernel", "", regex=False)
            .str.replace("_", " ")
        )

        plt.figure(figsize=figsize)

        palette_colors = sns.color_palette(palette, len(groups))

        xticks = []
        for i, group in enumerate(groups):
            group_df = df_filtered[df_filtered["kernel_func"].isin(group)]

            accuracy_range = (
                group_df.groupby("kernel_func")["accuracy"]
                .agg(["min", "max"])
                .reset_index()
            )
            accuracy_range["mean"] = (accuracy_range["min"] + accuracy_range["max"]) / 2
            accuracy_range["error"] = accuracy_range["max"] - accuracy_range["mean"]

            group_label = (
                group_names[i]
                if group_names and i < len(group_names)
                else f"Grupo {i+1}"
            )
            sns.scatterplot(
                data=accuracy_range,
                x="kernel_func",
                y="mean",
                color=palette_colors[i],
                marker="o",
                s=100,
                label=group_label,
            )

            plt.errorbar(
                x=accuracy_range["kernel_func"],
                y=accuracy_range["mean"],
                yerr=accuracy_range["error"],
                fmt="none",
                c=palette_colors[i],
                capsize=4,
                elinewidth=2,
                capthick=2,
            )

            xticks.extend(list(accuracy_range["kernel_func"]))
            xticks.append("")  # Ensure separation between groups

        # Remove the last empty string to ensure the last tick (separable) is visible
        xticks = xticks[:-1]

        plt.xticks(np.arange(len(xticks)), xticks, rotation=45, ha="right")

        if title:
            plt.title(title)

        plt.xlabel("Función de kernel")
        plt.ylabel("Precisión")

        plt.grid(True, which="both", linestyle="--", linewidth=0.5)

        if legend_title:
            plt.legend(title=legend_title, loc="lower right")
        else:
            plt.legend(loc="lower right")

        plt.tight_layout()
        plt.subplots_adjust(bottom=0.2)
        plt.show()

    def plot_dataset_evolution(
        self,
        datasets,
        classifiers,
        figsize=(7, 5),
        palette="RdBu",
        title=None,
        x_labels=None,
    ):
        if self.df is None:
            print("Dataframe is empty. Please load the CSV file first.")
            return

        df_filtered = self.df[
            self.df["experiment_name"].isin(datasets)
            & self.df["classifier"].isin(classifiers)
        ].copy()

        plt.figure(figsize=figsize)

        palette_colors = sns.color_palette(palette, len(classifiers))

        for i, classifier in enumerate(classifiers):
            classifier_df = df_filtered[df_filtered["classifier"] == classifier]

            evolution_df = (
                classifier_df.groupby("experiment_name")["accuracy"]
                .agg(["min", "max"])
                .reset_index()
            )
            evolution_df["mean"] = (evolution_df["min"] + evolution_df["max"]) / 2

            sns.lineplot(
                data=evolution_df,
                x="experiment_name",
                y="mean",
                color=palette_colors[i],
                label=classifier,
                linewidth=2,
            )
            plt.fill_between(
                evolution_df["experiment_name"],
                evolution_df["min"],
                evolution_df["max"],
                color=palette_colors[i],
                alpha=0.3,
            )

        plt.xlabel("Dataset")
        plt.ylabel("Accuracy")

        if x_labels:
            plt.xticks(np.arange(len(x_labels)), x_labels, rotation=45, ha="right")
        else:
            plt.xticks(rotation=45, ha="right")

        if title:
            plt.title(title)

        plt.grid(True, which="both", linestyle="--", linewidth=0.5)
        plt.legend(title="Classifier", loc="lower right")
        plt.tight_layout()
        plt.show()



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import ceil
import pennylane as qml
import numpy as np
import jax
from jax import numpy as jnp
import optax
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing import MinMaxScaler
from qcml.utils.model import train
from qcml.utils.model import chunk_vmapped_fn

jax.config.update("jax_enable_x64", True)


class DataReuploadingClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        n_layers=4,
        observable_type="single",
        convergence_interval=200,
        max_steps=10000,
        learning_rate=0.05,
        batch_size=32,
        max_vmap=None,
        jit=True,
        scaling=1.0,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax-jit"},
        random_state=42,
    ):
        r"""
        Data reuploading classifier from https://arxiv.org/abs/1907.02085. Here we code the 'multi-qubit classifier'.
        The model consists of layers of trainable data encoding unitaries. Each subsequent sequence of 3 elements of
        an input vector x are encoded into a SU(2) rotation via pennylane.Rot. The model therefore uses ceil(d/3)
        qubits.

        The cost function is given by eqn (23) in the plots, and includes trainable parameters for scaling input data
        (omegas), the class weights (alphas), and the circuit parameters (thetas). The cost
        function is a weighted sum of the squared differences between the ideal fidelity to the corresponding class state
        (here, either |0> or |1>) and the output state of the circuit, for a number of qubits specified by
        observable_weight. More specifically

        .. math::

            \ell(\vec{\theta},\vec{\omega},\vec{\alpha},\vec{x}_i) =
            \sum_{j=1}^{n_{\text{max}}}(\alpha_j^0F^0_j-(1-y_i))^2 + (\alpha_j^1F^1_j-y_i)^2

        where :math:`n_max` is given by observable_weight and :math:`F^0_j` is the fidelity of the jth output
        quibt to the state |0>.

        When observable_weight !=1 it is not clear from the plots how to predict labels. Here we look at the average
        fidelities to |0> and |1> of the qubits up to index self.observable_weight and take the label with the largest
        average fidelity. For self.observable_weight=1 this reverts to their strategy in the plots.

        Where possible, we have followed the numerical examples found in the repo which accompanies the plots:

        https://github.com/AdrianPerezSalinas/universal_qlassifier/tree/354d70f940ea737192de4063ff72f859c77e5760

        Args:
            n_layers (int): Number of blocks used in the trainable embedding. The data is uploaded n_layers+1 times.
            observable_type (str): Defines the number of qubits used to evaluate the weighed cost fucntion,
                either 'single', 'half' or 'full'.
                 max_steps (int): Maximum number of training steps. A warning will be raised if training did not
                    converge.
            learning_rate (float): Initial learning rate for training.
            convergence_interval (int): The number of loss values to consider to decide convergence.
            batch_size (int): Size of batches used for computing paraeeter updates.
            max_vmap (int or None): The maximum size of a chunk to vectorise over. Lower values use less memory.
                must divide batch_size.
            jit (bool): Whether to use just in time compilation.
            dev_type (str): string specifying the pennylane device type; e.g. 'default.qubit'.
            qnode_kwargs (str): the keyword arguments passed to the circuit qnode.
            scaling (float): Factor by which to scale the input data.
            random_state (int): Seed used for pseudorandom number generation.
        """

        # attributes that do not depend on data
        self.n_layers = n_layers
        self.observable_type = observable_type
        self.convergence_interval = convergence_interval
        self.max_steps = max_steps
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.dev_type = dev_type
        self.qnode_kwargs = qnode_kwargs
        self.jit = jit
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        if max_vmap is None:
            self.max_vmap = self.batch_size
        else:
            self.max_vmap = max_vmap

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None  # dictionary containing the trainable parameters
        self.n_qubits_ = None
        self.scaler = None  # data preprocessing
        self.circuit = None

    def generate_key(self):
        return jax.random.PRNGKey(self.rng.integers(100000))

    def construct_model(self):
        """Construct the quantum circuit used in the model."""

        dev = qml.device(self.dev_type, wires=self.n_qubits_)

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(params, x):
            """A variational quantum circuit with data reuploading.

            Args:
                params (array[float]): dictionary of parameters
                x (array[float]): input vector

            Returns:
                list: Expectation values of Pauli Z on each qubit
            """
            for layer in range(self.n_layers):
                x_idx = 0  # to keep track of the data index
                for i in range(self.n_qubits_):
                    # scaled inputs
                    angles = (
                        x[x_idx : x_idx + 3]
                        * params["omegas"][layer, x_idx : x_idx + 3]
                    )
                    qml.Rot(*angles, wires=i)

                    # variational
                    angles = params["thetas"][i, layer, :]
                    qml.Rot(*angles, wires=i)

                    x_idx += 3
                if layer % 2 == 0:
                    qml.broadcast(qml.CZ, range(self.n_qubits_), pattern="double")
                else:
                    qml.broadcast(qml.CZ, range(self.n_qubits_), pattern="double_odd")

            # final reupload without CZs
            x_idx = 0
            for i in range(self.n_qubits_):
                angles = (
                    x[x_idx : x_idx + 3]
                    * params["omegas"][self.n_layers, x_idx : x_idx + 3]
                    + params["thetas"][i, self.n_layers, :]
                )
                qml.Rot(*angles, wires=i)
                x_idx += 3

            return [qml.expval(qml.PauliZ(wires=[i])) for i in range(self.n_qubits_)]

        self.circuit = circuit

        def circuit_as_array(params, x):
            # pennylane returns a list in the circuit above, so we need this
            return jnp.array(circuit(params, x))

        if self.jit:
            circuit_as_array = jax.jit(circuit_as_array)
        self.forward = jax.vmap(circuit_as_array, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.
        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels that the classifier expects
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        self.n_features = n_features
        self.n_qubits_ = ceil(self.n_features / 3)

        if self.observable_type == "single":
            self.observable_weight = 1
        elif self.observable_type == "half":
            if self.n_qubits_ == 1:
                self.observable_weight = 1
            else:
                self.observable_weight = self.n_qubits_ // 2
        elif self.observable_type == "full":
            self.observable_weight = self.n_qubits_

        self.initialize_params()
        self.construct_model()

    def initialize_params(self):
        # This initialization is the same as found here:
        # https://github.com/AdrianPerezSalinas/universal_qlassifier/blob/master/problem_gen.py
        thetas = jax.random.uniform(
            self.generate_key(), shape=(self.n_qubits_, self.n_layers + 1, 3)
        )
        omegas = jax.random.uniform(
            self.generate_key(), shape=(self.n_layers + 1, self.n_qubits_ * 3)
        )
        alphas = jnp.ones(shape=(2, self.observable_weight))
        self.params_ = {"thetas": thetas, "omegas": omegas, "alphas": alphas}

    def fit(self, X, y):
        """Fit the model to data X and labels y.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """
        self.initialize(X.shape[1], classes=np.unique(y))

        self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
        self.scaler.fit(X)
        X = self.transform(X)

        def loss_fn(params, x, y):
            y_mat = jnp.vstack(
                [y for __ in range(self.observable_weight)]
            ).T  # repeated columns of labels
            y_mat0 = (1 - y_mat) / 2
            y_mat1 = (1 + y_mat) / 2
            alpha_mat0 = jnp.vstack(
                [params["alphas"][0, :] for __ in range(len(x))]
            )  # repeated rows of alpha parameters
            alpha_mat1 = jnp.vstack([params["alphas"][1, :] for __ in range(len(x))])

            expvals = self.forward(params, x)
            probs0 = (1 - expvals[:, : self.observable_weight]) / 2  # fidelity with |0>
            probs1 = (1 + expvals[:, : self.observable_weight]) / 2  # fidelity with |1>
            loss = (
                1
                / 2
                * (
                    jnp.sum(
                        (alpha_mat0 * probs0 - y_mat0) ** 2
                        + (alpha_mat1 * probs1 - y_mat1) ** 2
                    )
                )
            )  # eqn 23 in plots
            return loss / len(y)

        if self.jit:
            loss_fn = jax.jit(loss_fn)
        self.construct_model()
        self.initialize_params()
        optimizer = optax.adam

        self.params_ = train(
            self,
            loss_fn,
            optimizer,
            X,
            y,
            self.generate_key,
            convergence_interval=self.convergence_interval,
        )

        return self

    def predict(self, X):
        """Predict labels for batch of data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        check_is_fitted(self)
        predictions = self.predict_proba(X)
        mapped_predictions = np.argmax(predictions, axis=1)
        return np.take(self.classes_, mapped_predictions)

    def predict_proba(self, X):
        """Predict label probabilities for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """
        X = self.transform(X)
        predictions = self.chunked_forward(self.params_, X)
        predictions = jnp.mean(
            predictions[:, : self.observable_weight], axis=1
        )  # use mean of expvals over relevant qubits
        predictions_2d = np.c_[(1 - predictions) / 2, (1 + predictions) / 2]
        return predictions_2d

    def transform(self, X, preprocess=True):
        """
        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        if preprocess:
            if self.scaler is None:
                # if the model is unfitted, initialise the scaler here
                self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
                self.scaler.fit(X)

            X = self.scaler.transform(X)

        X = X * self.scaling
        X = jnp.pad(X, ((0, 0), (0, (3 - X.shape[1]) % 3)), "constant")
        return X


class DataReuploadingClassifierNoScaling(DataReuploadingClassifier):

    def construct_model(self):
        """Construct the quantum circuit used in the model."""

        dev = qml.device(self.dev_type, wires=self.n_qubits_)

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(params, x):
            """A variational quantum circuit with data reuploading.

            Args:
                params (array[float]): dictionary of parameters
                x (array[float]): input vector

            Returns:
                list: Expectation values of Pauli Z on each qubit
            """
            for layer in range(self.n_layers):
                x_idx = 0  # to keep track of the data index
                for i in range(self.n_qubits_):
                    angles = x[x_idx : x_idx + 3] + params["thetas"][i, layer, :]
                    qml.Rot(*angles, wires=i)
                    x_idx += 3
                if layer % 2 == 0:
                    qml.broadcast(qml.CZ, range(self.n_qubits_), pattern="double")
                else:
                    qml.broadcast(qml.CZ, range(self.n_qubits_), pattern="double_odd")

            # final reupload without CZs
            x_idx = 0
            for i in range(self.n_qubits_):
                angles = x[x_idx : x_idx + 3] + params["thetas"][i, self.n_layers, :]
                qml.Rot(*angles, wires=i)
                x_idx += 3

            return [qml.expval(qml.PauliZ(wires=[i])) for i in range(self.n_qubits_)]

        self.circuit = circuit

        def circuit_as_array(params, x):
            # pennylane returns a list in the circuit above, so we need this
            return jnp.array(circuit(params, x))

        if self.jit:
            circuit_as_array = jax.jit(circuit_as_array)
        self.forward = jax.vmap(circuit_as_array, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward

    def initialize_params(self):
        # This initialization is the same as found here:
        # https://github.com/AdrianPerezSalinas/universal_qlassifier/blob/master/problem_gen.py
        thetas = jax.random.uniform(
            self.generate_key(), shape=(self.n_qubits_, self.n_layers + 1, 3)
        )

        alphas = jnp.ones(shape=(2, self.observable_weight))
        self.params_ = {"thetas": thetas, "alphas": alphas}


class DataReuploadingClassifierNoTrainableEmbedding(DataReuploadingClassifier):

    def construct_model(self):
        """Construct the quantum circuit used in the model."""

        dev = qml.device(self.dev_type, wires=self.n_qubits_)

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(params, x):
            """A variational quantum circuit with data reuploading.

            Args:
                params (array[float]): dictionary of parameters
                x (array[float]): input vector

            Returns:
                list: Expectation values of Pauli Z on each qubit
            """
            for layer in range(self.n_layers):
                x_idx = 0  # to keep track of the data index
                for i in range(self.n_qubits_):
                    angles = (
                        x[x_idx : x_idx + 3]
                        * params["omegas"][layer, x_idx : x_idx + 3]
                    )
                    qml.Rot(*angles, wires=i)
                    x_idx += 3

            # final reupload without CZs
            x_idx = 0
            for i in range(self.n_qubits_):
                angles = (
                    x[x_idx : x_idx + 3]
                    * params["omegas"][self.n_layers, x_idx : x_idx + 3]
                )
                qml.Rot(*angles, wires=i)
                x_idx += 3

            for layer in range(self.n_layers):
                x_idx = 0  # to keep track of the data index
                for i in range(self.n_qubits_):
                    angles = params["thetas"][i, layer, :]
                    qml.Rot(*angles, wires=i)
                    x_idx += 3
                if layer % 2 == 0:
                    qml.broadcast(qml.CZ, range(self.n_qubits_), pattern="double")
                else:
                    qml.broadcast(qml.CZ, range(self.n_qubits_), pattern="double_odd")

                # final reupload without CZs
            x_idx = 0
            for i in range(self.n_qubits_):
                angles = params["thetas"][i, self.n_layers, :]
                qml.Rot(*angles, wires=i)
                x_idx += 3

            return [qml.expval(qml.PauliZ(wires=[i])) for i in range(self.n_qubits_)]

        self.circuit = circuit

        def circuit_as_array(params, x):
            # pennylane returns a list in the circuit above, so we need this
            return jnp.array(circuit(params, x))

        if self.jit:
            circuit_as_array = jax.jit(circuit_as_array)
        self.forward = jax.vmap(circuit_as_array, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward


class DataReuploadingClassifierNoCost(DataReuploadingClassifier):

    def fit(self, X, y):
        """Fit the model to data X and labels y.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """
        self.initialize(X.shape[1], classes=np.unique(y))

        self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
        self.scaler.fit(X)
        X = self.transform(X)

        def loss_fn(params, X, y):
            expvals = jnp.sum(self.forward(params, X), axis=1)
            probs = (1 - expvals * y) / 2  # the probs of incorrect classification
            return jnp.mean(probs)

        if self.jit:
            loss_fn = jax.jit(loss_fn)
        self.construct_model()
        self.initialize_params()
        optimizer = optax.adam

        self.params_ = train(
            self,
            loss_fn,
            optimizer,
            X,
            y,
            self.generate_key,
            convergence_interval=self.convergence_interval,
        )

        return self


class DataReuploadingClassifierSeparable(DataReuploadingClassifier):

    def construct_model(self):
        """Construct the quantum circuit used in the model."""

        dev = qml.device(self.dev_type, wires=self.n_qubits_)

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(params, x):
            """A variational quantum circuit with data reuploading.

            Args:
                params (array[float]): dictionary of parameters
                x (array[float]): input vector

            Returns:
                list: Expectation values of Pauli Z on each qubit
            """
            for layer in range(self.n_layers):
                x_idx = 0  # to keep track of the data index
                for i in range(self.n_qubits_):
                    angles = (
                        x[x_idx : x_idx + 3]
                        * params["omegas"][layer, x_idx : x_idx + 3]
                        + params["thetas"][i, layer, :]
                    )
                    qml.Rot(*angles, wires=i)
                    x_idx += 3

            # final reupload without CZs
            x_idx = 0
            for i in range(self.n_qubits_):
                angles = (
                    x[x_idx : x_idx + 3]
                    * params["omegas"][self.n_layers, x_idx : x_idx + 3]
                    + params["thetas"][i, self.n_layers, :]
                )
                qml.Rot(*angles, wires=i)
                x_idx += 3

            return [qml.expval(qml.PauliZ(wires=[i])) for i in range(self.n_qubits_)]

        self.circuit = circuit

        def circuit_as_array(params, x):
            # pennylane returns a list in the circuit above, so we need this
            return jnp.array(circuit(params, x))

        if self.jit:
            circuit_as_array = jax.jit(circuit_as_array)
        self.forward = jax.vmap(circuit_as_array, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcml.bench.kernel_grid import kernel_grid
from qcml.utils.kernel import kernel_transform
import logging

logger = logging.getLogger(__name__)


def get_kernel_transform():

    # Common parameters for transformations
    n_centers_options = [10, 20, 50, 100]
    scale_data_options = ["standardization", "minmax"]

    # Transformation parameter grids (renamed to transformation_params)
    transformation_params = [
        {},  # For None (no transformation)
    ]

    for kernel_entry in kernel_grid:
        kernel_func = kernel_entry["kernel_func"]
        kernel_params = kernel_entry["kernel_params"]

        # Create a transformation grid for the current kernel
        transformation_entry = {
            "kernel_func": [kernel_func],
            "kernel_params": [kernel_params],
            "n_centers": n_centers_options,
            "scale_data": scale_data_options,
        }

        # Add the entry to the transformation_params list
        transformation_params.append(transformation_entry)

    # Generate transformation_func
    transformation_func = [None] + [kernel_transform] * (len(transformation_params) - 1)
    logger.debug(
        f"Getting kernel transformation with {len(transformation_func)} combinations"
    )

    return transformation_func, transformation_params



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split

logger = logging.getLogger(__name__)


def load_mnist_data():
    # Load MNIST dataset from sklearn
    mnist = fetch_openml("mnist_784", version=1, as_frame=False)

    # Separate features and labels
    X = mnist.data
    y = mnist.target.astype(int)

    return X, y


def split_data(X, y, split_ratio):
    train_ratio, val_ratio = map(float, split_ratio.split("/"))
    train_ratio /= 100.0
    val_ratio /= 100.0
    if train_ratio + val_ratio != 1.0:
        raise ValueError("Split ratio must sum to 100 (e.g., '80/20').")
    X_train, X_val, y_train, y_val = train_test_split(
        X, y, test_size=val_ratio, random_state=42
    )
    logger.debug(
        f"Data split into training samples: {len(y_train)}, validation samples: {len(y_val)}"
    )
    return X_train, X_val, y_train, y_val


def validate_input_data(
    X=None,
    y=None,
    X_train=None,
    y_train=None,
    X_val=None,
    y_val=None,
    split_ratio="80/20",
):
    if (X is not None and y is None) or (X is None and y is not None):
        raise ValueError("Both X and y must be provided if using whole dataset.")

    if (X_train is not None and y_train is None) or (
        X_train is None and y_train is not None
    ):
        raise ValueError(
            "Both X_train and y_train must be provided if using pre-split dataset."
        )

    if (X_val is not None and y_val is None) or (X_val is None and y_val is not None):
        raise ValueError(
            "Both X_val and y_val must be provided if using pre-split dataset."
        )

    if X is not None and split_ratio is None:
        raise ValueError("Split ratio must be provided if using whole dataset.")

    if X is not None and y is not None:
        X_train, X_val, y_train, y_val = split_data(X, y, split_ratio)
        logger.info(
            f"Data split into training and validation sets using split ratio: {split_ratio}"
        )
    elif X_train is None or y_train is None or X_val is None or y_val is None:
        raise ValueError(
            "Insufficient data provided. Provide either whole dataset or pre-split dataset."
        )

    return X_train, X_val, y_train, y_val



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pandas as pd
import numpy as np
import logging

# Configure logging
logging.basicConfig(level=logging.DEBUG)


def get_dataset(
    dataset_name,
    root_path="datasets/",
    file_type="join",
    parameters=None,
    split_ratio=None,
):
    """
    Read data from CSV files where each row is a data sample.
    The columns are the input features, and the last column specifies a label.

    If file_type is 'join', the function checks for both 'train' and 'test' files. If both are found, it concatenates the data.
    If file_type is 'split', the function will return X_train, X_test, y_train, y_test.
    If just one file exists, the data will be split according to split_ratio.

    Args:
        dataset_name (str): Name of the dataset. This is a required parameter.
        root_path (str, optional): Root directory of the datasets. Defaults to 'datasets/'.
        file_type (str, optional): Type of the file handling, either 'join', 'split', 'train', or 'test'. Defaults to 'join'.
        parameters (list, optional): Additional parameters to append to the file name, delimited by '_'. Defaults to None.
        split_ratio (str, optional): Split ratio for train/test split, e.g., '80/20'. Only used if file_type is 'split'.

    Returns:
        X (ndarray): 2-d array of inputs.
        y (ndarray): Array of labels.
        If file_type is 'split', returns X_train, X_test, y_train, y_test.
    """
    if parameters:
        params = "_".join(parameters)
    else:
        params = ""

    def load_file(file_type=None):
        # Build the filename based on the dataset name, file_type, and additional parameters
        file_name = f"{dataset_name}"
        if params:
            file_name = f"{file_name}_{params}"
        if file_type:
            file_name = f"{file_name}_{file_type}"
        file_name = f"{file_name}.csv"

        # Construct the full file path
        file_path = os.path.join(root_path, dataset_name, file_name)

        try:
            logging.debug(f"Attempting to load file: {file_path}")
            # Attempt to read the CSV file
            data = pd.read_csv(file_path, header=None)
            X = data.iloc[:, :-1].values
            y = data.iloc[:, -1].values
            return X, y
        except Exception as e:
            # Log any other exceptions that may occur
            logging.error(f"An error occurred while reading the file: {file_path}")
            logging.error(e)
            return None, None

    if file_type == "join":
        logging.debug("File type is 'join'. Checking for both train and test files.")
        # Check if both train and test files exist
        X_train, y_train = load_file("train")
        X_test, y_test = load_file("test")

        if X_train is not None and X_test is not None:
            logging.debug("Both train and test files found. Concatenating data.")
            # If both train and test files exist, concatenate them
            X = np.vstack((X_train, X_test))
            y = np.hstack((y_train, y_test))
            return X, y
        elif X_train is not None:
            logging.debug("Only train file found. Returning train data.")
            return X_train, y_train
        elif X_test is not None:
            logging.debug("Only test file found. Returning test data.")
            return X_test, y_test
        else:
            logging.debug(
                "Neither train nor test files found. Trying without file_type."
            )
            # If neither train nor test files are found, try loading without file_type
            return load_file(None)

    elif file_type == "split":
        logging.debug("File type is 'split'. Checking for train and test files.")
        # Check if both train and test files exist
        X_train, y_train = load_file("train")
        X_test, y_test = load_file("test")

        if X_train is not None and X_test is not None:
            logging.debug("Both train and test files found. Returning separate data.")
            return X_train, X_test, y_train, y_test
        else:
            logging.debug(
                "Either train or test file is missing. Attempting to split from a single file."
            )
            # If only one file exists, perform the split based on split_ratio
            X, y = load_file(None)
            if X is not None and split_ratio is not None:
                train_ratio, test_ratio = map(float, split_ratio.split("/"))
                train_size = int(len(X) * (train_ratio / (train_ratio + test_ratio)))
                X_train, X_test = X[:train_size], X[train_size:]
                y_train, y_test = y[:train_size], y[train_size:]
                logging.debug(f"Data split with ratio {split_ratio}.")
                return X_train, X_test, y_train, y_test
            else:
                logging.error(
                    "Failed to load data for splitting or invalid split ratio provided."
                )
                return None, None, None, None

    else:
        logging.debug(f"File type is '{file_type}'. Loading specified type.")
        # Load only the specified file type ('train' or 'test')
        return load_file(file_type)



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from sklearn.preprocessing import StandardScaler


def perceptron(x, w, b):
    """Transforms inputs according to a perceptron.
    Args:
        x (ndarray): input of shape (dim_hyperplanes,)
        w (ndarray): input-to-hidden weight matrix of shape (dim_hyperplanes,)
        b (float): bias
    """
    if np.dot(w, x) + b > 0:
        return 1
    return 0


def predict(x, weights, biases):
    """Implements the parity prediction logic.

    Args:
        x (ndarray): input of shape (dim_hyperplanes,)
        weights (ndarray): array of weight vectors defining
            the orientation of the hyperplanes
        biases (ndarray): array of biases defining
            the offset of the hyperplanes
    """
    preds = [perceptron(x, w, b) for w, b in zip(weights, biases)]
    n_ones = np.sum(preds)
    if n_ones % 2 == 0:
        return 1
    else:
        return -1


def generate_hyperplanes_parity(n_samples, n_features, n_hyperplanes, dim_hyperplanes):
    """Data generation procedure for 'hyperplanes and parity'.

    Args:
        n_samples (int): number of samples to generate
        n_features (int): dimension of the data samples
        n_hyperplanes (int): number of hyperplanes to use for prediction
        dim_hyperplanes (int): dimension of space in which
            hyperplanes are defined
    """

    # define hyperplanes
    weights = np.random.uniform(size=(n_hyperplanes, dim_hyperplanes))
    biases = np.random.uniform(size=(n_hyperplanes,))

    # hack: initially create more data than we need,
    # and then subselect to get balanced classes
    X = np.random.normal(size=(4 * n_samples, dim_hyperplanes))
    y = np.array([predict(x, weights, biases) for x in X])
    A = X[y == 1]
    B = X[y == -1]
    assert len(A) >= n_samples // 2
    assert len(B) >= n_samples // 2
    X = np.r_[A[: n_samples // 2], B[: n_samples // 2]]
    y = np.array([-1] * (n_samples // 2) + [1] * (n_samples // 2))

    # embed into feature space by a linear transform
    emb_W = np.random.uniform(size=(dim_hyperplanes, n_features))
    X = X @ emb_W

    s = StandardScaler()
    X = s.fit_transform(X)

    return X, y



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def generate_linearly_separable(n_samples, n_features, margin):
    """Data generation procedure for 'linearly separable'.

    Args:
        n_samples (int): number of samples to generate
        n_features (int): dimension of the data samples
        margin (float): width between hyperplane and closest samples
    """

    w_true = np.ones(n_features)

    # hack: sample more data than we need randomly from a hypercube
    X = 2 * np.random.rand(2 * n_samples, n_features) - 1

    # only retain data outside a margin
    X = [x for x in X if np.abs(np.dot(x, w_true)) > margin]
    X = X[:n_samples]

    y = [np.dot(x, w_true) for x in X]
    y = [-1 if y_ > 0 else 1 for y_ in y]
    return X, y



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from sklearn.decomposition import PCA
from matplotlib.colors import ListedColormap, rgb_to_hsv, hsv_to_rgb


def palette_saturated(palette_name="coolwarm", n_colors=2, midpoint=0.2):
    palette = sns.color_palette(palette_name, n_colors=256)
    start_idx = int(midpoint * 256)
    end_idx = int((1 - midpoint) * 256)
    selected_colors = (
        palette[start_idx : start_idx + n_colors // 2]
        + palette[end_idx - n_colors // 2 : end_idx]
    )
    return selected_colors


def adjust_saturation(colors, factor=1.5):
    """Adjust the saturation of a list of colors."""
    colors = np.array(colors)
    hsv_colors = rgb_to_hsv(colors[:, :3])
    hsv_colors[:, 1] = np.clip(hsv_colors[:, 1] * factor, 0, 1)
    rgb_colors = hsv_to_rgb(hsv_colors)
    return rgb_colors



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import logging

logger = logging.getLogger(__name__)


def get_gpu_info():
    try:
        gpus = jax.devices("gpu")
        num_gpus = len(gpus)
        gpu_names = [gpu.device_kind for gpu in gpus]
        return num_gpus, gpu_names
    except Exception as e:
        logger.error(f"Error retrieving GPU information: {e}")
        return 0, []



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

model_grid = {
    "LinearSVC": {
        "C": [0.1, 1, 10, 100],  # Regularization parameter
    },
    "SVC": {
        "C": [0.1, 1, 10, 100],  # Regularization parameter
        "kernel": ["precomputed"],
    },
    "KernelPerceptron": {
        "n_iter": [-1],  # Number of iterations
        "max_iter": [2000],  # Maximum number of iterations
        "kernel": ["precomputed"],
    },
    "Perceptron": {
        "eta0": [0.01, 0.1, 0.5, 1.0],  # Learning rate
    },
    "PerceptronCustom": {
        "n_iter": [1000],  # Number of iterations
    },
    "MLPClassifier": {
        "hidden_layer_sizes": [
            (50,),
            (100,),
            (100, 50),
            (100, 100),
        ],  # Configurations for hidden layers
        "learning_rate_init": [
            0.001,
            0.01,
        ],  # Initial learning rate (specific to MLPClassifier)
        "max_iter": [3000],  # Maximum number of iterations
    },
    "MLPClassifierCustom": {
        "hidden_layer_sizes": [
            (50,),
            (100,),
            (100, 50),
            (100, 100),
        ],  # Configurations for hidden layers
        "learning_rate": [
            0.001,
            0.01,
        ],  # Learning rate for optimizer (specific to MLPClassifierCustom)
        "max_iter": [2000],  # Maximum number of iterations
        "num_classes": [2],  # Number of output classes
    },
    "KernelMLPClassifier": {
        "hidden_layer_sizes": [
            ("k", 50),
            ("k", 100),
            ("k", 100, 50),
            ("k", 100, 100),
        ],  # Configurations for hidden layers
        "learning_rate_init": [
            0.001,
            0.01,
        ],  # Initial learning rate (specific to MLPClassifier)
        "max_iter": [2000],  # Maximum number of iterations
        "num_classes": [2],  # Number of output classes
    },
    "QuanvolutionalNeuralNetwork": {
        "max_vmap": [32],
        "batch_size": [32],
        "learning_rate": [0.0001, 0.001, 0.01],
        "n_qchannels": [1, 5, 10],
        "qkernel_shape": [2, 3],
        "kernel_shape": [2, 3, 5],
    },
}



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pennylane as qml
import numpy as np
import jax
from sklearn.base import BaseEstimator, ClassifierMixin
from qcml.utils.model import *

jax.config.update("jax_enable_x64", True)


class CircuitCentricClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        n_input_copies=2,
        n_layers=4,
        convergence_interval=200,
        max_steps=10000,
        learning_rate=0.001,
        batch_size=32,
        max_vmap=None,
        jit=True,
        scaling=1.0,
        random_state=42,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax-jit"},
    ):
        r"""
        A variational quantum model based on https://arxiv.org/pdf/1804.03680v2.pdf.

        Uses amplitude encoding as the data embedding, but implements several
        copies of the initial state. The data vector :math:`x` of size :math:`N`
        gets padded to the next power of 2 by values :math:`1/len(\tilde{x})`
        (so that the new data vector :math:`\tilde{x}` has length :math:`2^n > N`).
        The padded vector gets normalised to :math:`\bar{\tilde{x}}`, and
        embedded into the amplitudes of several copies of :math:`n`-qubit
        quantum states. Altogether, the variational circuit acts on a quantum
        state of amplitudes :math:`\bar{\tilde{x}} \otimes \bar{\tilde{x}} \otimes ...`.

        The total number of qubits :math:`d\lceil \log(n_features) \rceil` of
        this model depends on the number of features :math:`N` and the number
        of copies :math:`d`.

        The variational part of the circuit uses general single qubit rotations
        as trainable gates. Each layer in the ansatz first applies a rotation
        to each qubit, and then controlled rotations connecting each qubit
        :math:`i` to qubit :math:`i+r \text{mod } n`, where :math:`r` repeatedly runs
        through the range :math:`[0,...,n-1]`. The number of layers in the
        ansatz is a hyperparameter of the model.

        The result of the model is the sigma-z expectation of the first qubit,
        plus a trainable scalar bias. Training is via the square loss.

        Args:
            n_input_copies (int): Number of copies of the amplitude embedded
                state to produce.
            n_layers (int): Number of layers in the variational ansatz.
            convergence_interval (int): The number of loss values to consider to decide convergence.
            max_steps (int): Maximum number of training steps. A warning will
                be raised if training did not converge.
            learning_rate (float): Initial learning rate for gradient descent.
            batch_size (int): Size of batches used for computing parameter updates.
            max_vmap (int or None): The maximum size of a chunk to vectorise over. Lower values use less memory.
                must divide batch_size.
            jit (bool): Whether to use just in time compilation.
            dev_type (str): Pennylane device type; e.g. 'default.qubit.jax'.
            qnode_kwargs (str): Keyword arguments for the circuit qnode.
            scaling (float): Factor by which to scale the input data.
            random_state (int): Seed used for pseudorandom number generation.
        """

        # attributes that do not depend on data
        self.n_input_copies = n_input_copies
        self.n_layers = n_layers
        self.convergence_interval = convergence_interval
        self.max_steps = max_steps
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.dev_type = dev_type
        self.qnode_kwargs = qnode_kwargs
        self.jit = jit
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        if max_vmap is None:
            self.max_vmap = self.batch_size
        else:
            self.max_vmap = max_vmap

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None  # Dictionary containing the trainable parameters
        self.n_qubits_ = None
        self.scaler = None
        self.circuit = None

    def generate_key(self) -> jax.Array:
        """
        Generates a random key used in sampling batches.

        Returns:
            jax.Array: _description_
        """
        return jax.random.PRNGKey(self.rng.integers(1000000))

    def construct_model(self):

        dev = qml.device(self.dev_type, wires=self.n_qubits_)

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(params, x):
            # classically compute the copies of inputs
            # since AmplitudeEmbedding cannot be called
            # multiple times on jax device
            tensor_x = x
            for i in range(self.n_input_copies - 1):
                tensor_x = jnp.kron(tensor_x, x)

            qml.AmplitudeEmbedding(tensor_x, wires=range(self.n_qubits_))
            qml.StronglyEntanglingLayers(params["weights"], wires=range(self.n_qubits_))
            return qml.expval(qml.PauliZ(0))

        self.circuit = circuit

        if self.jit:
            circuit = jax.jit(circuit)

        def circuit_plus_bias(params, x):
            return circuit(params, x) + params["b"]

        if self.jit:
            circuit_plus_bias = jax.jit(circuit_plus_bias)
        self.forward = jax.vmap(circuit_plus_bias, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.
        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        n_qubits_per_copy = int(np.ceil(np.log2(n_features)))
        self.n_qubits_ = self.n_input_copies * n_qubits_per_copy

        self.construct_model()
        self.initialize_params()

    def initialize_params(self):
        # initialise the trainable parameters
        shape = qml.StronglyEntanglingLayers.shape(
            n_layers=self.n_layers, n_wires=self.n_qubits_
        )
        weights = jax.random.uniform(
            self.generate_key(), minval=0, maxval=2 * np.pi, shape=shape
        )
        b = jnp.array(0.01)
        self.params_ = {"weights": weights, "b": b}

    def fit(self, X, y):
        """Fit the model to data X and labels y.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """

        self.initialize(n_features=X.shape[1], classes=np.unique(y))

        X = self.transform(X)

        optimizer = optax.adam

        def loss_fn(params, X, y):
            pred = self.forward(
                params, X
            )  # jnp.stack([self.forward(params, x) for x in X])
            return jnp.mean(optax.l2_loss(pred, y))

        if self.jit:
            loss_fn = jax.jit(loss_fn)
        self.params_ = train(
            self,
            loss_fn,
            optimizer,
            X,
            y,
            self.generate_key,
            convergence_interval=self.convergence_interval,
        )

        return self

    def predict(self, X):
        """Predict labels for batch of data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        predictions = self.predict_proba(X)
        mapped_predictions = np.argmax(predictions, axis=1)
        return np.take(self.classes_, mapped_predictions)

    def predict_proba(self, X):
        """Predict label probabilities for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """
        X = self.transform(X)
        predictions = self.chunked_forward(self.params_, X)
        predictions_2d = np.c_[(1 - predictions) / 2, (1 + predictions) / 2]
        return predictions_2d

    def transform(self, X, preprocess=False):
        """
        The feature vectors padded to the next power of 2 and then normalised.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        n_features = X.shape[1]
        X = X * self.scaling

        n_qubits_per_copy = int(np.ceil(np.log2(n_features)))
        max_n_features = 2**n_qubits_per_copy
        n_padding = max_n_features - n_features
        padding = np.ones(shape=(len(X), n_padding)) / max_n_features

        X_padded = np.c_[X, padding]
        X_normalised = np.divide(
            X_padded, np.expand_dims(np.linalg.norm(X_padded, axis=1), axis=1)
        )
        return X_normalised



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from sklearn.decomposition import PCA
from matplotlib.colors import ListedColormap, rgb_to_hsv, hsv_to_rgb
from qic.utils.visualization.seaborn import palette_saturated, adjust_saturation


def plot_decision_boundary(X, y, model, ax, palette, saturation_factor=1.5):
    x_min, x_max = X[:, 0].min() - 1, X[:, 0].max() + 1
    y_min, y_max = X[:, 1].min() - 1, X[:, 1].max() + 1
    xx, yy = np.meshgrid(np.linspace(x_min, x_max, 500), np.linspace(y_min, y_max, 500))

    Z = model.predict(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)

    # Adjust saturation of the palette
    saturated_palette = adjust_saturation(palette, factor=saturation_factor)

    cmap = ListedColormap(saturated_palette)

    ax.contourf(xx, yy, Z, alpha=0.3, levels=np.arange(y.max() + 2) - 0.5, cmap=cmap)


def plot_2d_dataset(
    train_csv,
    test_csv,
    feature1_index=0,
    feature2_index=1,
    axis_titles=None,
    palette=None,
    fig_size=(4, 3),
    decision_boundary=None,
    saturation_factor=1.5,
):
    df_train = pd.read_csv(train_csv, header=None)
    df_test = pd.read_csv(train_csv, header=None)

    df_combined = pd.concat([df_train, df_test], ignore_index=True)

    class_col = df_combined.columns[-1]
    features = df_combined.iloc[:, :-1]

    if features.shape[1] > 2:
        pca = PCA(n_components=2)
        features_pca = pca.fit_transform(features)
        df_combined = pd.DataFrame(features_pca, columns=["PCA1", "PCA2"])
        df_combined["Class"] = pd.concat(
            [df_train.iloc[:, -1], df_test.iloc[:, -1]], ignore_index=True
        )
    else:
        df_combined = pd.concat([features, df_combined[class_col]], axis=1)
        df_combined.columns = ["Feature1", "Feature2", "Class"]

    X = df_combined.iloc[:, :2].values
    y = df_combined["Class"].values

    if palette is None:
        unique_classes = np.unique(y).size
        palette = palette_saturated(palette_name="coolwarm", n_colors=unique_classes)

    fig, ax = plt.subplots(figsize=fig_size)

    # Plot decision boundary if a classifier is provided
    if decision_boundary is not None:
        decision_boundary.fit(X, y)
        plot_decision_boundary(
            X, y, decision_boundary, ax, palette, saturation_factor=saturation_factor
        )

    # Plot the data points
    sns.scatterplot(x=X[:, 0], y=X[:, 1], hue=y, palette=palette, ax=ax, edgecolor="k")

    if axis_titles:
        ax.set_xlabel(axis_titles[0])
        ax.set_ylabel(axis_titles[1])
    else:
        ax.set_xlabel("")
        ax.set_ylabel("")

    ax.legend().remove()  # No legend by default

    plt.show()



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import pennylane as qml
import numpy as np
import jax

jax.config.update("jax_enable_x64", True)
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.extmath import softmax
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import MinMaxScaler
from qcml.utils.model import chunk_vmapped_fn


class QuantumKitchenSinks(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        linear_model=LogisticRegression(penalty=None, solver="lbfgs", tol=10e-4),
        n_episodes=100,
        n_qfeatures="full",
        var=1.0,
        jit=True,
        max_vmap=None,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax", "diff_method": None},
        scaling=1.0,
        random_state=42,
    ):
        r"""
        Quantum kitchen sinks model classical data classification.

        Based on: https://arxiv.org/pdf/1806.08321.pdf

        The quantum computer is used to generate random feature vectors which are
        fed to a linear classifier that we implement using scikit-learn's LogisticRegression (note logistic
        regression is a standard method to perform linear classification despite the name)

        The feature map procedure works as follows:

        1. Linearly transform an input feature vector :math:`x` of length :math:`n`
        via :math:`x' = \omega x + \beta` using random :math:`\omega, \beta`.
        The output is of shape `(n_episodes, n_qfeatures)`.

        2. Feed each row vector of the matrix :math:`x'` into a quantum circuit that
        returns `n_qfeatures` samples. The samples are concatenated in a feature vector :math:`x`
        of length `n_episodes*n_qfeatures`, which is the input to the linear model.

        .. note::

            It is not stated in the plots how to generalise the circuits to higher qubit numbers;
            this implementation is a simple generalisation of the circuit in Fig 2(c), which consists
            of encoding the features into single qubits via X rotaiton gates, and perfoming a sequence
            of CNOT gates on nearest neighbour and next-nearest neighbour qubits.

        Args:
            linear_model (sklearn Estimator): linear model to use with the transformed features
            n_episodes (int): Number of features fed into the linear model after data transformation.
            n_qfeatures (str, int): Determines the number of features fed into the quantum circuit to transform the
                data. This is the number of qubits used by the model. If 'full', the number of qubits is equal
                to the number of input features, if 'half' it is half the number of input features.
            var (float): detemined the variance of the matrix `\omega` used to lienarly transform the input
                features.
            max_vmap (int or None): The maximum size of a chunk to vectorise over. Lower values use less memory.
                must divide batch_size.
            jit (bool): Whether to use just in time compilation.
            dev_type (str): string specifying the pennylane device type; e.g. 'default.qubit'
            qnode_kwargs (str): the key word arguments passed to the circuit qnode.
            random_state (int): Seed used for pseudorandom number generation.
        """

        # attributes that do not depend on data
        self.linear_model = linear_model
        self.n_episodes = n_episodes
        self.n_qfeatures = n_qfeatures
        self.var = var
        self.jit = jit
        self.dev_type = dev_type
        self.qnode_kwargs = qnode_kwargs
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        if max_vmap is None:
            self.max_vmap = 100000
        else:
            self.max_vmap = max_vmap

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None
        self.n_qubits_ = None
        self.scaler = None  # data scaler for inputs
        self.scaler2 = None  # data scaler for quantum generated features
        self.circuit = None

    def generate_key(self):
        return jax.random.PRNGKey(self.rng.integers(1000000))

    def construct_model(self):
        """
        The circuit used to generate the feature vectors
        It is not clear from the plots how to generalise this to higher qubit numbers.
        This is a simple generalisation of the circuit in Fig 2(c).
        """

        pattern = [[i, i + 2] for i in range(0, self.n_qubits_ - 2)]

        dev = qml.device(
            self.dev_type,
            wires=self.n_qubits_,
            shots=1,
            prng_key=self.generate_key(),
        )

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(Q):

            for i, q in enumerate(Q):
                qml.RX(q, wires=i)
            qml.broadcast(qml.CNOT, wires=range(self.n_qubits_), pattern="double")
            qml.broadcast(qml.CNOT, wires=range(self.n_qubits_), pattern=pattern)

            return qml.sample(wires=range(self.n_qubits_))

        self.circuit = circuit

        if self.jit:
            circuit = jax.jit(circuit)
        circuit = chunk_vmapped_fn(jax.vmap(circuit), 0, self.max_vmap)

        self.forward = circuit

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.

        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels that the classifier expects
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        self.n_features_ = n_features

        if self.n_qfeatures == "full":
            self.n_qubits_ = n_features
        elif self.n_qfeatures == "half":
            self.n_qubits_ = int(np.ceil(n_features / 2))
        else:
            self.n_qubits_ = int(self.n_qfeatures)

        self.construct_model()
        self.initialize_params()

    def initialize_params(self):
        # initialise the parameters
        omegas = jax.random.normal(
            key=self.generate_key(),
            shape=(self.n_episodes, self.n_qubits_, self.n_features_),
        ) * np.sqrt(self.var)
        betas = (
            2
            * np.pi
            * jax.random.uniform(
                key=self.generate_key(), shape=(self.n_episodes, self.n_qubits_)
            )
        )
        self.params_ = {"omegas": np.array(omegas), "betas": np.array(betas)}

    def fit(self, X, y):
        """Fit the model to data X and labels y.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels 1-, 1 of shape (n_samples,)
        """

        self.linear_model.random_state = self.rng.integers(100000)

        self.initialize(X.shape[1], np.unique(y))

        self.scaler = MinMaxScaler(feature_range=(-1, 1))
        self.scaler.fit(X)
        features = self.transform(X)

        start = time.time()
        self.linear_model.fit(features, y)
        end = time.time()
        self.params_["weights"] = self.linear_model.coef_
        self.training_time_ = end - start
        return self

    def predict(self, X):
        """Predict labels for batch of data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        predictions = self.predict_proba(X)
        mapped_predictions = np.argmax(predictions, axis=1)
        return np.take(self.classes_, mapped_predictions)

    def predict_proba(self, X):
        """Predict label probabilities for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """
        X = self.transform(X)
        predictions_2d = self.linear_model.predict_proba(X)
        return softmax(predictions_2d, copy=False)

    def transform(self, X, preprocess=True):
        """
        Apply the feature map: The inputs go through a random linear transformation
        followed by a quantum circuit.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        if (
            self.params_["betas"] is None
            or self.params_["omegas"] is None
            or self.circuit is None
        ):
            raise ValueError("Model cannot predict without fitting to data first.")

        if preprocess:
            if self.scaler is None:
                # if the model is unfitted, initialise the scaler here
                self.scaler = MinMaxScaler(feature_range=(-1, 1))
                self.scaler.fit(X)
            X = self.scaler.transform(X)
        X = X * self.scaling

        n_data = X.shape[0]
        input_features = np.zeros([self.n_episodes, n_data, self.n_qubits_])
        for e in range(self.n_episodes):
            stacked_beta = np.stack([self.params_["betas"][e] for __ in range(n_data)])
            input_features[e] = (self.params_["omegas"][e] @ X.T + stacked_beta.T).T
        input_features = np.reshape(input_features, (n_data * self.n_episodes, -1))

        features = self.forward(input_features)
        features = np.reshape(features, (self.n_episodes, n_data, -1))
        features = np.array([features[:, i, :] for i in range(n_data)])
        features = np.reshape(features, (n_data, -1))

        if self.scaler2 is None:
            self.scaler2 = StandardScaler().fit(features)

        return features



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.base import BaseEstimator, ClassifierMixin
import pennylane as qml
import numpy as np
from qcml.utils.model import *
from sklearn.svm import SVC
from sklearn.preprocessing import MinMaxScaler
from qcml.models.classical import KernelPerceptron

jax.config.update("jax_enable_x64", True)


class SeparableVariationalClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        encoding_layers=1,
        learning_rate=0.001,
        batch_size=32,
        max_vmap=None,
        jit=True,
        max_steps=10000,
        random_state=42,
        scaling=1.0,
        convergence_interval=200,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax"},
    ):
        r"""
        Variational model that uses only separable operations (i.e. there is no entanglement in the model). The circuit
        consists of layers of encoding gates and parameterised unitaries followed by measurement of an observable.

        Each encoding layer consists of a trainiable arbitrary qubit rotation on each qubit followed by
        a product angle embedding of the input data, using RY gates. A final layer of trainable qubit rotations is
        applied at the end of the circuit.

        The obserable O is the mean value of Pauli Z observables on each of the output qubits. The value of this
        observable is used to predict the probability for class 1 as :math:`P(+1)=\sigma(6\langle O \rangle)`
        where :math`\sigma` is the logistic funciton. The model is then fit using the cross entropy loss.

        Args:
            encoding_layers (int): number of layers in the data encoding circuit.
            learning_rate (float): learning rate for gradient descent.
            batch_size (int): Size of batches used for computing parameter updates.
            max_vmap (int or None): The maximum size of a chunk to vectorise over. Lower values use less memory.
                must divide batch_size.
            jit (bool): Whether to use just in time compilation.
            convergence_interval (int): The number of loss values to consider to decide convergence.
            max_steps (int): Maximum number of training steps. A warning will be raised if training did not converge.
            dev_type (str): string specifying the pennylane device type; e.g. 'default.qubit'.
            qnode_kwargs (str): the keyword arguments passed to the circuit qnode.
            random_state (int): Seed used for pseudorandom number generation
        """
        # attributes that do not depend on data
        self.encoding_layers = encoding_layers
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.max_steps = max_steps
        self.dev_type = dev_type
        self.qnode_kwargs = qnode_kwargs
        self.jit = jit
        self.convergence_interval = convergence_interval
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        if max_vmap is None:
            self.max_vmap = self.batch_size
        else:
            self.max_vmap = max_vmap

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None  # Dictionary containing the trainable parameters
        self.n_qubits_ = None
        self.scaler = None  # data scaler will be fitted on training data
        self.circuit = None

    def generate_key(self):
        return jax.random.PRNGKey(self.rng.integers(1000000))

    def construct_model(self):

        dev = qml.device(self.dev_type, wires=1)

        @qml.qnode(dev, **self.qnode_kwargs)
        def single_qubit_circuit(weights, x):
            """
            To avoid using large state vectors we define the circuit for a single qubit and combine
            below to get the full circuit
            """
            for layer in range(self.encoding_layers):
                qml.Rot(
                    weights[layer, 0], weights[layer, 1], weights[layer, 2], wires=0
                )
                qml.RY(x, wires=0)
            qml.Rot(
                weights[self.encoding_layers, 0],
                weights[self.encoding_layers, 1],
                weights[self.encoding_layers, 2],
                wires=0,
            )
            return qml.expval(qml.PauliZ(wires=0))

        self.circuit = single_qubit_circuit

        def circuit(params, x):
            return jnp.mean(
                jnp.array(
                    [
                        single_qubit_circuit(params["weights"][i], x[i])
                        for i in range(self.n_qubits_)
                    ]
                )
            )

        if self.jit:
            circuit = jax.jit(circuit)
        self.forward = jax.vmap(circuit, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.

        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels that the classifier expects
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        self.n_qubits_ = n_features
        self.initialize_params()
        self.construct_model()

    def initialize_params(self):
        # initialise the trainable parameters
        weights = (
            2
            * jnp.pi
            * jax.random.uniform(
                shape=(self.n_qubits_, self.encoding_layers + 1, 3),
                key=self.generate_key(),
            )
        )

        self.params_ = {"weights": weights}

    def fit(self, X, y):
        """Fit the model to data X and labels y.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """
        self.initialize(X.shape[1], np.unique(y))

        self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
        self.scaler.fit(X)
        X = self.transform(X)

        optimizer = optax.adam

        def loss_fn(params, X, y):
            # we multiply by 6 because a relevant domain of the sigmoid function is [-6,6]
            vals = self.forward(params, X) * 6
            y = jax.nn.relu(y)  # convert to 0,1
            return jnp.mean(optax.sigmoid_binary_cross_entropy(vals, y))

        if self.jit:
            loss_fn = jax.jit(loss_fn)
        self.params_ = train(
            self,
            loss_fn,
            optimizer,
            X,
            y,
            self.generate_key,
            convergence_interval=self.convergence_interval,
        )

        return self

    def predict(self, X):
        """Predict labels for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred (np.ndarray): Predicted labels of shape (n_samples,)
        """
        predictions = self.predict_proba(X)
        mapped_predictions = np.argmax(predictions, axis=1)
        return np.take(self.classes_, mapped_predictions)

    def predict_proba(self, X):
        """Predict label probabilities for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """
        X = self.transform(X)
        predictions = self.chunked_forward(self.params_, X)
        predictions_2d = np.c_[(1 - predictions) / 2, (1 + predictions) / 2]
        return predictions_2d

    def transform(self, X, preprocess=True):
        """
        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        if preprocess:
            if self.scaler is None:
                self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
                self.scaler.fit(X)
            X = self.scaler.transform(X)
        return X * self.scaling


class SeparableKernelClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        encoding_layers=1,
        classifier=None,
        C=1.0,
        jit=True,
        random_state=42,
        scaling=1.0,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax", "diff_method": None},
    ):
        r"""
        A kernel model that uses a separable embedding. The embedding consists of layers of fixed single qubit X
        rotations by an angle :math:`pi/4` on the Bloch sphere and single qubit Y rotation data encoding gates.

        The kernel is given by

        .. math::
            k(x,x')=\vert\langle 0 \vert U^\dagger(x')U(x)\vert 0 \rangle\vert^2


        Args:
            encoding_layers (int): number of layers in the data encoding circuit.
            svm (sklearn.svm.SVC): scikit-learn SVC class object used to fit the model from the kernel matrix.
            C (float): regularization parameter for the SVC. Lower values imply stronger regularization.
            jit (bool): Whether to use just in time compilation.
            random_state (int): Seed used for pseudorandom number generation.
            dev_type (str): string specifying the pennylane device type; e.g. 'default.qubit'.
            qnode_kwargs (str): the keyword arguments passed to the circuit qnode.
            scaling (float): Factor by which to scale the input data.
        """
        # attributes that do not depend on data
        self.encoding_layers = encoding_layers
        self.dev_type = dev_type
        self.qnode_kwargs = qnode_kwargs
        self.jit = jit
        self.classifier = classifier
        self.C = C
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None  # Dictionary containing the trainable parameters
        self.n_qubits_ = None
        self.scaler = None  # data scaler will be fitted on training data
        self.circuit = None

    def generate_key(self):
        return jax.random.PRNGKey(self.rng.integers(1000000))

    def construct_circuit(self):
        projector = np.zeros([2, 2])
        projector[0, 0] = 1.0

        dev = qml.device(self.dev_type, wires=1)

        @qml.qnode(dev, **self.qnode_kwargs)
        def qubit_circuit(x):
            """
            Compute the overlap probability for a single qubit to avoid large state vectors.
            This is used to compute the full overlap below.
            """
            for layer in range(self.encoding_layers):
                qml.RX(-np.pi / 4, wires=0)
                qml.RY(-x[0], wires=0)
            for layer in range(self.encoding_layers):
                qml.RY(x[1], wires=0)
                qml.RX(np.pi / 4, wires=0)
            return qml.expval(qml.Hermitian(projector, wires=0))

        self.circuit = qubit_circuit

        def circuit(x):
            probs = [
                qubit_circuit(jnp.array([x[i], x[i + self.n_qubits_]]))
                for i in range(self.n_qubits_)
            ]
            return jnp.prod(jnp.array(probs))

        if self.jit:
            circuit = jax.jit(circuit)
        self.forward = circuit

        return self.forward

    def precompute_kernel(self, X1, X2):
        """
        compute the kernel matrix relative to data sets X1 and X2
        Args:
            X1 (np.array): first dataset of input vectors
            X2 (np.array): second dataset of input vectors
        Returns:
            kernel_matrix (np.array): matrix of size (len(X1),len(X2)) with elements K(x_1,x_2)
        """
        dim1 = len(X1)
        dim2 = len(X2)

        # concatenate all pairs of vectors
        Z = np.array(
            [np.concatenate((X1[i], X2[j])) for i in range(dim1) for j in range(dim2)]
        )
        self.construct_circuit()
        kernel_values = [self.forward(z) for z in Z]
        # reshape the values into the kernel matrix
        kernel_matrix = np.reshape(kernel_values, (dim1, dim2))

        return kernel_matrix

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.

        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels that the classifier expects
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        self.n_qubits_ = n_features
        self.params_ = {}
        self.construct_circuit()

    def fit(self, X, y):
        """Fit the model to data X and labels y. Uses sklearn's SVM classifier

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """

        self.svm.random_state = int(
            jax.random.randint(
                self.generate_key(), shape=(1,), minval=0, maxval=1000000
            )
        )

        self.initialize(X.shape[1], np.unique(y))

        self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
        self.scaler.fit(X)
        X = self.transform(X)

        self.params_ = {"x_train": X}
        kernel_matrix = self.precompute_kernel(X, X)

        start = time.time()
        # we are updating this value here, in case it was
        # changed after initialising the model
        if isinstance(self.classifier, SVC):
            self.svm.C = self.C
        self.svm.fit(kernel_matrix, y)
        end = time.time()
        self.training_time_ = end - start

        return self

    def predict(self, X):
        """Predict labels for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred (np.ndarray): Predicted labels of shape (n_samples,)
        """
        X = self.transform(X)
        kernel_matrix = self.precompute_kernel(X, self.params_["x_train"])
        return self.svm.predict(kernel_matrix)

    def predict_proba(self, X):
        """Predict label probabilities for data X.
        note that this may be inconsistent with predict; see the sklearn docummentation for details.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """
        if "x_train" not in self.params_:
            raise ValueError("Model cannot predict without fitting to data first.")

        X = self.transform(X)
        kernel_matrix = self.precompute_kernel(X, self.params_["x_train"])
        return self.svm.predict_proba(kernel_matrix)

    def transform(self, X, preprocess=True):
        """
        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        if preprocess:
            if self.scaler is None:
                # if the model is unfitted, initialise the scaler here
                self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
                self.scaler.fit(X)
            X = self.scaler.transform(X)
        return X * self.scaling



import logging
import os
import json


def dask_setup(worker):
    # Read environment variables
    log_level = os.getenv("QCML_DASK_LEVEL", "INFO").upper()
    output = os.getenv("QCML_DASK_OUTPUT", "both")
    logs_path = os.getenv("QCML_DASK_LOGS_PATH", "logs/")
    log_filename = os.getenv("QCML_DASK_LOG_FILENAME", "dask_worker.log")
    max_bytes = int(os.getenv("QCML_DASK_MAX_BYTES", "1048576"))
    backup_count = int(os.getenv("QCML_DASK_BACKUP_COUNT", "3"))
    terminal_level = os.getenv("QCML_DASK_TERMINAL_LEVEL", log_level).upper()
    file_level = os.getenv("QCML_DASK_FILE_LEVEL", log_level).upper()
    log_format = os.getenv(
        "QCML_DASK_LOG_FORMAT",
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)",
    )
    use_json = os.getenv("QCML_DASK_USE_JSON", "False") == "True"
    use_color = os.getenv("QCML_DASK_USE_COLOR", "True") == "True"
    asynchronous = os.getenv("QCML_DASK_ASYNCHRONOUS", "False") == "True"
    add_context = os.getenv("QCML_DASK_ADD_CONTEXT", "False") == "True"
    slack_notify = os.getenv("QCML_DASK_SLACK_NOTIFY", "False") == "True"

    slack_token = os.getenv("QCML_DASK_SLACK_TOKEN")
    slack_channel = os.getenv("QCML_DASK_SLACK_CHANNEL")
    hide_logs = os.getenv("QCML_DASK_HIDE_LOGS")
    keyword_filters = os.getenv("QCML_DASK_KEYWORD_FILTERS")
    context_info = os.getenv("QCML_DASK_CONTEXT_INFO")

    if context_info:
        context_info = json.loads(context_info)

    if hide_logs:
        hide_logs = hide_logs.split(",")

    if keyword_filters:
        keyword_filters = keyword_filters.split(",")

    # Configure logging
    logger = logging.getLogger()
    logger.setLevel(logging.NOTSET)

    # Remove any existing handlers
    if logger.hasHandlers():
        logger.handlers.clear()

    # Setup formatter
    formatter = logging.Formatter(log_format)

    # Add handlers based on output configuration
    if output in ("terminal", "both"):
        console_handler = logging.StreamHandler()
        console_handler.setLevel(getattr(logging, terminal_level))
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    if output in ("file", "both"):
        if not os.path.exists(logs_path):
            os.makedirs(logs_path)

        file_handler = logging.handlers.RotatingFileHandler(
            os.path.join(logs_path, log_filename),
            maxBytes=max_bytes,
            backupCount=backup_count,
        )
        file_handler.setLevel(getattr(logging, file_level))
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # Slack handler setup (if needed)
    if slack_notify and slack_token and slack_channel:
        try:
            from slack_sdk import WebClient
            from slack_sdk.errors import SlackApiError

            class SlackHandler(logging.Handler):
                def __init__(self, slack_client, channel):
                    super().__init__(level=logging.INFO)
                    self.slack_client = slack_client
                    self.channel = channel

                def emit(self, record):
                    log_entry = self.format(record)
                    try:
                        self.slack_client.chat_postMessage(
                            channel=self.channel, text=f"```{log_entry}```"
                        )
                    except SlackApiError as e:
                        logging.error(
                            f"Failed to send log to Slack: {e.response['error']}"
                        )

            slack_client = WebClient(token=slack_token)
            slack_handler = SlackHandler(slack_client, slack_channel)
            slack_handler.setFormatter(formatter)
            logger.addHandler(slack_handler)

        except ImportError:
            logging.error("Slack SDK not installed. Slack notifications will not work.")

    # Set logging levels for specified libraries
    for library in ["pennylane", "jax", "distributed", "matplotlib"]:
        logging.getLogger(library).setLevel(logging.INFO)
        logger.info(f"{library} logs are set to INFO level.")

    logging.info("Dask worker logging setup complete.")
    worker.qml_devices = {}



# Copyright 2024 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run hyperparameter search and store results with a command-line script."""

import numpy as np
import sys
import os
import time
import argparse
import logging

logging.getLogger().setLevel(logging.INFO)
from importlib import import_module
import pandas as pd
from pathlib import Path
import matplotlib.pyplot as plt
from sklearn.model_selection import GridSearchCV
from qcml.utils.hyperparam_search import (
    read_data,
    construct_hyperparameter_grid,
)
from qcml.bench.hyperparameter_settings import hyper_parameter_settings

np.random.seed(42)

logging.info("cpu count:" + str(os.cpu_count()))


if __name__ == "__main__":
    # Create an argument parser
    parser = argparse.ArgumentParser(
        description="Run experiments with hyperparameter search.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--classifier-name",
        help="Classifier to run",
    )

    parser.add_argument(
        "--dataset-path",
        help="Path to the dataset",
    )

    parser.add_argument(
        "--results-path", default=".", help="Path to store the experiment results"
    )

    parser.add_argument(
        "--clean",
        help="True or False. Remove previous results if it exists",
        dest="clean",
        default=False,
        type=bool,
    )

    parser.add_argument(
        "--hyperparameter-scoring",
        type=list,
        nargs="+",
        default=["accuracy", "roc_auc"],
        help="Scoring for hyperparameter search.",
    )

    parser.add_argument(
        "--hyperparameter-refit",
        type=str,
        default="accuracy",
        help="Refit scoring for hyperparameter search.",
    )

    parser.add_argument(
        "--plot-loss",
        help="True or False. Plot loss history for single fit",
        dest="plot_loss",
        default=False,
        type=bool,
    )

    parser.add_argument(
        "--n-jobs", type=int, default=-1, help="Number of parallel threads to run"
    )

    # Parse the arguments along with any extra arguments that might be model specific
    args, unknown_args = parser.parse_known_args()

    if any(arg is None for arg in [args.classifier_name, args.dataset_path]):
        msg = "\n================================================================================"
        msg += "\nA classifier from qml.benchmarks.model and dataset path are required. E.g., \n \n"
        msg += "python run_hyperparameter_search \ \n--classifier DataReuploadingClassifier \ \n--dataset-path train.csv\n"
        msg += "\nCheck all arguments for the script with \n"
        msg += "python run_hyperparameter_search --help\n"
        msg += "================================================================================"
        raise ValueError(msg)

    # Add model specific arguments to override the default hyperparameter grid
    hyperparam_grid = construct_hyperparameter_grid(
        hyper_parameter_settings, args.classifier_name
    )
    for hyperparam in hyperparam_grid:
        hp_type = type(hyperparam_grid[hyperparam][0])
        parser.add_argument(
            f"--{hyperparam}",
            type=hp_type,
            nargs="+",
            default=hyperparam_grid[hyperparam],
            help=f"{hyperparam} grid values for {args.classifier_name}",
        )

    args = parser.parse_args(unknown_args, namespace=args)

    for hyperparam in hyperparam_grid:
        override = getattr(args, hyperparam)
        if override is not None:
            hyperparam_grid[hyperparam] = override
    logging.info(
        "Running hyperparameter search experiment with the following settings\n"
    )
    logging.info(args.classifier_name)
    logging.info(args.dataset_path)
    logging.info(" ".join(args.hyperparameter_scoring))
    logging.info(args.hyperparameter_refit)
    logging.info(
        "Hyperparam grid:"
        + " ".join(
            [
                (str(key) + str(":") + str(hyperparam_grid[key]))
                for key in hyperparam_grid.keys()
            ]
        )
    )

    experiment_path = args.results_path
    results_path = os.path.join(experiment_path, "results")

    if not os.path.exists(results_path):
        os.makedirs(results_path)

    ###################################################################
    # Get the classifier, dataset and search methods from the arguments
    ###################################################################
    Classifier = getattr(import_module("qml_benchmarks.models"), args.classifier_name)
    classifier_name = Classifier.__name__

    # Run the experiments save the results
    train_dataset_filename = os.path.join(args.dataset_path)
    X, y = read_data(train_dataset_filename)

    dataset_path_obj = Path(args.dataset_path)
    results_filename_stem = " ".join(
        [Classifier.__name__ + "_" + dataset_path_obj.stem + "_GridSearchCV"]
    )

    # If we have already run this experiment then continue
    if os.path.isfile(os.path.join(results_path, results_filename_stem + ".csv")):
        if args.clean is False:
            msg = "\n================================================================================="
            msg += "\nResults exist in " + os.path.join(
                results_path, results_filename_stem + ".csv"
            )
            msg += "\nSpecify --clean True to override results or new --results-path"
            msg += "\n================================================================================="
            logging.warning(msg)
            sys.exit(msg)
        else:
            logging.warning(
                "Cleaning existing results for ",
                os.path.join(results_path, results_filename_stem + ".csv"),
            )

    ###########################################################################
    # Single fit to check everything works
    ###########################################################################
    classifier = Classifier()
    a = time.time()
    classifier.fit(X, y)
    b = time.time()
    acc_train = classifier.score(X, y)
    logging.info(
        " ".join(
            [
                classifier_name,
                "Dataset path",
                args.dataset_path,
                "Train acc:",
                str(acc_train),
                "Time single run",
                str(b - a),
            ]
        )
    )
    if hasattr(classifier, "loss_history_"):
        if args.plot_loss:
            plt.plot(classifier.loss_history_)
            plt.xlabel("Iterations")
            plt.ylabel("Loss")
            plt.show()

    if hasattr(classifier, "n_qubits_"):
        logging.info(" ".join(["Num qubits", f"{classifier.n_qubits_}"]))

    ###########################################################################
    # Hyperparameter search
    ###########################################################################
    gs = GridSearchCV(
        estimator=classifier,
        param_grid=hyperparam_grid,
        scoring=args.hyperparameter_scoring,
        refit=args.hyperparameter_refit,
        verbose=3,
        n_jobs=-1,
    ).fit(X, y)
    logging.info("Best hyperparams")
    logging.info(gs.best_params_)

    df = pd.DataFrame.from_dict(gs.cv_results_)
    df.to_csv(os.path.join(results_path, results_filename_stem + ".csv"))

    best_df = pd.DataFrame(
        list(gs.best_params_.items()), columns=["hyperparameter", "best_value"]
    )

    # Save best hyperparameters to a CSV file
    best_df.to_csv(
        os.path.join(results_path, results_filename_stem + "-best-hyperparameters.csv"),
        index=False,
    )



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate datasets for the HIDDEN MANIFOLD and HIDDEN MANIFOLD DIFF benchmarks."""

import os
import numpy as np
from sklearn.model_selection import train_test_split
from qcml.data import generate_hidden_manifold_model

np.random.seed(3)
os.makedirs("hidden_manifold", exist_ok=True)

manifold_dimension = 6
n_samples = 300

for n_features in range(2, 21):
    X, y = generate_hidden_manifold_model(n_samples, n_features, manifold_dimension)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2)

    name_train = f"hidden_manifold/hidden_manifold-{manifold_dimension}manifold-{n_features}d_train.csv"
    data_train = np.c_[X_train, y_train]
    np.savetxt(name_train, data_train, delimiter=",")

    name_test = f"hidden_manifold/hidden_manifold-{manifold_dimension}manifold-{n_features}d_test.csv"
    data_test = np.c_[X_test, y_test]
    np.savetxt(name_test, data_test, delimiter=",")


os.makedirs("hidden_manifold_diff", exist_ok=True)

n_features = 10
n_samples = 300

for manifold_dimension in range(2, 21):
    X, y = generate_hidden_manifold_model(n_samples, n_features, manifold_dimension)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2)

    name_train = f"hidden_manifold_diff/hidden_manifold-10d-{manifold_dimension}manifold_train.csv"
    data_train = np.c_[X_train, y_train]
    np.savetxt(name_train, data_train, delimiter=",")

    name_test = f"hidden_manifold_diff/hidden_manifold-10d-{manifold_dimension}manifold_test.csv"
    data_test = np.c_[X_test, y_test]
    np.savetxt(name_test, data_test, delimiter=",")



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
from jax import vmap
import jax
import numpy as np
import logging
import time
import concurrent.futures
from sklearn.preprocessing import StandardScaler, MinMaxScaler

logger = logging.getLogger(__name__)


def kernel_transform(
    X_train,
    X_val,
    y_train,
    y_val,
    kernel_func,
    kernel_params=None,
    n_centers=10,
    scale_data=None,
):
    """
    Apply a kernel transformation to the training and validation data.

    Parameters:
    - X_train: Training data, shape (n_train_samples, n_features).
    - X_val: Validation data, shape (n_val_samples, n_features).
    - kernel_func: Kernel function to be applied.
    - kernel_params: Additional parameters for the kernel function.
    - n_centers: Number of centers to choose from X_train.
    - scale_data: 'standardization' or 'minmax' to choose the scaling method, or None for no scaling.

    Returns:
    - X_train_transformed: Transformed training data, shape (n_train_samples, n_centers).
    - X_val_transformed: Transformed validation data, shape (n_val_samples, n_centers).
    """
    # Optionally scale the data
    if scale_data == "standardization":
        scaler = StandardScaler()
    elif scale_data == "minmax":
        scaler = MinMaxScaler()
    else:
        scaler = None

    if scaler is not None:
        X_train = scaler.fit_transform(X_train)
        X_val = scaler.transform(X_val)

    logger.debug(
        f"Transforming data with kernel: {kernel_func.__name__}, parameters: {kernel_params}, and scaling with: {scale_data}"
    )
    # Choose random centers from X_train
    rng = np.random.default_rng()
    indices = rng.choice(X_train.shape[0], size=n_centers, replace=False)
    centers = X_train[indices]

    # Ensure kernel_params is a dictionary
    if kernel_params is None:
        kernel_params = {}

    # Compute the kernel matrix for X_train and centers
    X_train_transformed = compute_gram_matrix(
        X_train, centers, kernel_func, **kernel_params
    )

    # Compute the kernel matrix for X_val and centers
    X_val_transformed = compute_gram_matrix(
        X_val, centers, kernel_func, **kernel_params
    )
    logger.debug(
        f"Transformation done: {kernel_func.__name__}, parameters: {kernel_params}, and scaling with: {scale_data}"
    )
    return X_train_transformed, X_val_transformed, y_train, y_val


def jitted_gram_matrix_batched(X1, X2, kernel_func, batch_size=100, **kernel_params):
    start_time = time.time()

    n_samples_X1 = X1.shape[0]
    n_samples_X2 = X2.shape[0]
    Gram_matrix = jnp.zeros((n_samples_X1, n_samples_X2))

    def wrapped_kernel_func(x1, x2):
        return kernel_func(x1, x2, **kernel_params)

    # Use vmap without JIT
    kernel_func_vmap = jax.vmap(
        jax.vmap(wrapped_kernel_func, in_axes=(None, 0)), in_axes=(0, None)
    )

    device_info = jax.devices()[0]

    logger.debug(
        f"Vmap applied for Gram matrix computation on device: {device_info.platform.upper()}. "
        f"X1 shape: {X1.shape}, X2 shape: {X2.shape}. "
        f"Kernel: {kernel_func.__name__}. "
        f"Expected output shape: ({X1.shape[0]}, {X2.shape[0]})"
    )

    # Batched computation without JIT
    for i in range(0, n_samples_X1, batch_size):
        for j in range(0, n_samples_X2, batch_size):
            X1_batch = X1[i : i + batch_size]
            X2_batch = X2[j : j + batch_size]

            # Compute the kernel in a batched manner
            batch_result = kernel_func_vmap(X1_batch, X2_batch)
            jax.block_until_ready(
                batch_result
            )  # Ensures computation is done and memory is cleared
            Gram_matrix = Gram_matrix.at[i : i + batch_size, j : j + batch_size].set(
                batch_result
            )

    execution_time = time.time() - start_time

    logger.debug(
        f"Execution time for batched Gram matrix computation: {execution_time:.4f}s. "
        f"Real output shape: {Gram_matrix.shape}"
    )

    return Gram_matrix


def jitted_gram_matrix(X1, X2, kernel_func, **kernel_params):
    start_time = time.time()

    # Create a wrapper to pass kernel_params to kernel_func
    def wrapped_kernel_func(x1, x2):
        return kernel_func(x1, x2, **kernel_params)

    # Vectorized and JIT-compiled Gram matrix computation
    kernel_func_vmap = jax.jit(
        jax.vmap(jax.vmap(wrapped_kernel_func, in_axes=(None, 0)), in_axes=(0, None))
    )

    # Trigger JIT compilation by calling the function with dummy data
    _ = kernel_func_vmap(X1[:1], X2[:1])
    compilation_time = time.time() - start_time

    device_info = jax.devices()[0]

    logger.debug(
        f"JIT and vmap applied for Gram matrix computation on device: {device_info.platform.upper()}. "
        f"Compilation time: {compilation_time:.4f}s. "
        f"X1 shape: {X1.shape}, X2 shape: {X2.shape}. "
        f"Kernel: {kernel_func.__name__}, JIT enabled: {jax.config.read('jax_enable_x64')}. "
        f"Expected output shape: ({X1.shape[0]}, {X2.shape[0]})"
    )

    # Measure execution time of the actual Gram matrix computation
    execution_start_time = time.time()
    gram_matrix = kernel_func_vmap(X1, X2)
    execution_time = time.time() - execution_start_time

    logger.debug(
        f"Execution time for Gram matrix computation: {execution_time:.4f}s. "
        f"Real output shape: {gram_matrix.shape}"
    )

    return gram_matrix


def compute_kernel_parallel(X1, X2, kernel_func, **kernel_params):
    n_samples_X1 = X1.shape[0]
    n_samples_X2 = X2.shape[0]
    Gram_matrix = jnp.zeros((n_samples_X1, n_samples_X2))

    def compute_entry(i, j):
        return kernel_func(X1[i], X2[j], **kernel_params)

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = {
            executor.submit(compute_entry, i, j): (i, j)
            for i in range(n_samples_X1)
            for j in range(n_samples_X2)
        }
        logger.debug(
            f"Computing Gram matrix for kernel function: {kernel_func.__name__}, kernel_params: {kernel_params}, input sizes: {X1.shape}, {X2.shape}, split in {len(futures)}"
        )
        for future in concurrent.futures.as_completed(futures):
            i, j = futures[future]
            Gram_matrix = Gram_matrix.at[i, j].set(future.result())

    return Gram_matrix


def batched_gram_matrix(X1, X2, kernel_func, batch_size=100, **kernel_params):
    n_samples_X1 = X1.shape[0]
    n_samples_X2 = X2.shape[0]
    Gram_matrix = jnp.zeros((n_samples_X1, n_samples_X2))

    for i in range(0, n_samples_X1, batch_size):
        X1_batch = X1[i : i + batch_size]
        for j in range(0, n_samples_X2, batch_size):
            X2_batch = X2[j : j + batch_size]

            # Vectorizing the kernel function application over batches
            batch_result = vmap(
                lambda x1: vmap(lambda x2: kernel_func(x1, x2, **kernel_params))(
                    X2_batch
                )
            )(X1_batch)

            Gram_matrix = Gram_matrix.at[i : i + batch_size, j : j + batch_size].set(
                batch_result
            )

    return Gram_matrix


def compute_gram_matrix(X1, X2, kernel_func, **kernel_params):
    Gram_matrix = batched_gram_matrix(
        X1, X2, kernel_func, batch_size=100, **kernel_params
    )

    # Log details about the computation after the matrix has been computed
    logger.debug(
        f"Computed Gram matrix with kernel function: {kernel_func.__name__}, "
        f"kernel_params: {kernel_params}, input sizes: {X1.shape}, {X2.shape}, "
        f"output size: {Gram_matrix.shape}"
    )

    return Gram_matrix


def gram_matrix(X1, X2, kernel_func, **kernel_params):
    start_time = time.time()

    # Expand the dimensions to perform broadcasting over the entire matrix
    X1_expanded = X1[:, None, :]  # Shape: (n_samples_X1, 1, features)
    X2_expanded = X2[None, :, :]  # Shape: (1, n_samples_X2, features)

    # Apply the kernel function to the expanded arrays, leveraging broadcasting
    Gram_matrix = jax.vmap(
        lambda x1: jax.vmap(lambda x2: kernel_func(x1, x2, **kernel_params))(X2)
    )(X1)

    jax.block_until_ready(
        Gram_matrix
    )  # Ensures the computation is done before measuring time
    execution_time = time.time() - start_time
    logger.debug(f"Execution time for Gram matrix computation: {execution_time:.4f}s")

    return Gram_matrix



import matplotlib.pyplot as plt
import numpy as np


class CircleNotation:
    """
    CircleNotation class for creating annotated circles on a matplotlib Axes.

    Parameters:
    - label (list): List of labels for each circle.
    - inner_color (list): List of colors for the inner circles.
    - outer_color (list): List of colors for the outer circles.
    - angle (list): List of angles for drawing lines.
    - x_offset (list): List of x-coordinates for circle centers.
    - y_offset (list): List of y-coordinates for circle centers.
    - len_outer_circle (list): List of radii for outer circles.
    - len_inner_circle (list): List of radii for inner circles.
    - line_length (float, optional): Length of lines drawn from circles. Default is DEFAULT_LINE_LENGTH.
    - line_color (str, optional): Color of the lines. Default is DEFAULT_LINE_COLOR.
    - omit_line (list, optional): List of booleans indicating whether to omit lines for each circle. Default is DEFAULT_OMIT_LINE.

    Attributes:
    - DEFAULT_LINE_LENGTH (float): Default length of lines drawn from circles.
    - DEFAULT_LINE_COLOR (str): Default color of the lines.
    - DEFAULT_CIRCLE_COLOR (str): Default color of circles.
    - DEFAULT_TEXT_COLOR (str): Default color of text.
    - DEFAULT_HEAD_WIDTH (float): Default width of arrowhead in lines.
    - DEFAULT_HEAD_LENGTH (float): Default length of arrowhead in lines.
    - DEFAULT_FONT_SIZE (int): Default font size of text.
    - DEFAULT_TEXT_DISPLACEMENT (float): Default vertical displacement of text.
    - DEFAULT_OMIT_LINE (bool): Default value indicating whether to omit lines for each circle.
    - DEFAULT_CIRCLE_FILL (bool): Default value indicating whether circles should be filled.
    - DEFAULT_X_OFFSET (float): Default x-coordinate for circle centers.
    - DEFAULT_Y_OFFSET (float): Default y-coordinate for circle centers.
    - DEFAULT_CIRCLE_RADIUS (float): Default radius of circles.
    - DEFAULT_ANGLE (float): Default angle for drawing lines.
    - DEFAULT_TEXT_HA (str): Default horizontal alignment of text.
    - DEFAULT_TEXT_VA (str): Default vertical alignment of text.
    """

    DEFAULT_LINE_LENGTH = 1
    DEFAULT_LINE_COLOR = "black"
    DEFAULT_CIRCLE_COLOR = "black"
    DEFAULT_TEXT_COLOR = "black"
    DEFAULT_HEAD_WIDTH = 0
    DEFAULT_HEAD_LENGTH = 0
    DEFAULT_FONT_SIZE = 12
    DEFAULT_TEXT_DISPLACEMENT = -1.35
    DEFAULT_OMIT_LINE = False
    DEFAULT_CIRCLE_FILL = False
    DEFAULT_X_OFFSET = 0
    DEFAULT_Y_OFFSET = 0
    DEFAULT_CIRCLE_RADIUS = 1
    DEFAULT_ANGLE = 0
    DEFAULT_TEXT_HA = "center"
    DEFAULT_TEXT_VA = "center"

    def __init__(
        self,
        label,
        inner_color,
        outer_color,
        angle,
        x_offset,
        y_offset,
        len_outer_circle,
        len_inner_circle,
        line_length=DEFAULT_LINE_LENGTH,
        line_color=DEFAULT_LINE_COLOR,
        omit_line=DEFAULT_OMIT_LINE,
    ):
        """
        Initialize CircleNotation instance.

        Parameters:
        - label (list): List of labels for each circle.
        - inner_color (list): List of colors for the inner circles.
        - outer_color (list): List of colors for the outer circles.
        - angle (list): List of angles for drawing lines.
        - x_offset (list): List of x-coordinates for circle centers.
        - y_offset (list): List of y-coordinates for circle centers.
        - len_outer_circle (list): List of radii for outer circles.
        - len_inner_circle (list): List of radii for inner circles.
        - line_length (float, optional): Length of lines drawn from circles. Default is DEFAULT_LINE_LENGTH.
        - line_color (str, optional): Color of the lines. Default is DEFAULT_LINE_COLOR.
        - omit_line (list, optional): List of booleans indicating whether to omit lines for each circle. Default is DEFAULT_OMIT_LINE.
        """

        self.fig, self.ax = plt.subplots()
        self.label = label
        self.inner_color = inner_color
        self.outer_color = outer_color
        self.line_color = line_color
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.len_outer_circle = len_outer_circle
        self.len_inner_circle = len_inner_circle
        self.line_length = line_length
        self.angle = angle
        self.omit_line = omit_line

    def draw_circle(
        self,
        x_offset=DEFAULT_X_OFFSET,
        y_offset=DEFAULT_Y_OFFSET,
        radius=DEFAULT_CIRCLE_RADIUS,
        color=DEFAULT_CIRCLE_COLOR,
        is_filled=DEFAULT_CIRCLE_FILL,
    ):
        """
        Draw a circle on the Axes.

        Parameters:
        - x_offset (float, optional): x-coordinate of the circle center. Default is DEFAULT_X_OFFSET.
        - y_offset (float, optional): y-coordinate of the circle center. Default is DEFAULT_Y_OFFSET.
        - radius (float, optional): Radius of the circle. Default is DEFAULT_CIRCLE_RADIUS.
        - color (str, optional): Color of the circle. Default is DEFAULT_CIRCLE_COLOR.
        - is_filled (bool, optional): Whether the circle should be filled. Default is DEFAULT_CIRCLE_FILL.
        """
        circle = plt.Circle((x_offset, y_offset), radius, color=color, fill=is_filled)
        self.ax.add_patch(circle)

    def draw_text(
        self,
        label,
        x_offset=DEFAULT_X_OFFSET,
        y_offset=DEFAULT_Y_OFFSET,
        text_displacement=DEFAULT_TEXT_DISPLACEMENT,
        color=DEFAULT_TEXT_COLOR,
        fontsize=DEFAULT_FONT_SIZE,
        ha=DEFAULT_TEXT_HA,
        va=DEFAULT_TEXT_VA,
    ):
        """
        Draw text on the Axes.

        Parameters:
        - label (str): Text label to be drawn.
        - x_offset (float, optional): x-coordinate of the text. Default is DEFAULT_X_OFFSET.
        - y_offset (float, optional): y-coordinate of the text. Default is DEFAULT_Y_OFFSET.
        - text_displacement (float, optional): Vertical displacement of the text. Default is DEFAULT_TEXT_DISPLACEMENT.
        - color (str, optional): Color of the text. Default is DEFAULT_TEXT_COLOR.
        - fontsize (int, optional): Font size of the text. Default is DEFAULT_FONT_SIZE.
        - ha (str, optional): Horizontal alignment of the text. Default is DEFAULT_TEXT_HA.
        - va (str, optional): Vertical alignment of the text. Default is DEFAULT_TEXT_VA.
        """
        self.ax.text(
            x_offset,
            y_offset + text_displacement,
            label,
            color=color,
            ha=ha,
            va=va,
            fontsize=fontsize,
        )

    def draw_line(
        self,
        angle=DEFAULT_ANGLE,
        x_offset=DEFAULT_X_OFFSET,
        y_offset=DEFAULT_Y_OFFSET,
        color=DEFAULT_LINE_COLOR,
        head_width=DEFAULT_HEAD_WIDTH,
        head_length=DEFAULT_HEAD_LENGTH,
    ):
        """
        Draw a line on the Axes.

        Parameters:
        - angle (float, optional): Angle of the line. Default is DEFAULT_ANGLE.
        - x_offset (float, optional): x-coordinate of the line start. Default is DEFAULT_X_OFFSET.
        - y_offset (float, optional): y-coordinate of the line start. Default is DEFAULT_Y_OFFSET.
        - color (str, optional): Color of the line. Default is DEFAULT_LINE_COLOR.
        - head_width (float, optional): Width of the arrowhead. Default is DEFAULT_HEAD_WIDTH.
        - head_length (float, optional): Length of the arrowhead. Default is DEFAULT_HEAD_LENGTH.
        """
        line_x = self.line_length * np.cos(angle) - head_length
        line_y = self.line_length * np.sin(angle)

        self.ax.arrow(
            x_offset,
            y_offset,
            line_x,
            line_y,
            head_width=head_width,
            head_length=head_length,
            fc=color,
            ec=color,
        )

    def draw_single(self, i):
        """
        Draw a single set of circles, text, and lines.

        Parameters:
        - i (int): Index of the circle set to be drawn.
        """
        self.draw_circle(
            x_offset=self.x_offset[i],
            y_offset=self.y_offset[i],
            radius=self.len_outer_circle[i],
            color=self.outer_color[i],
        )
        self.draw_circle(
            x_offset=self.x_offset[i],
            y_offset=self.y_offset[i],
            radius=self.len_inner_circle[i],
            color=self.inner_color[i],
            is_filled=True,
        )
        self.draw_text(
            self.label[i], x_offset=self.x_offset[i], y_offset=self.y_offset[i]
        )
        if not self.omit_line[i]:
            self.draw_line(
                angle=self.angle[i],
                x_offset=self.x_offset[i],
                y_offset=self.y_offset[i],
            )

    def draw_all(self):
        """Draw all circle sets."""
        for i in range(len(self.angle)):
            self.draw_single(i)

        # Set aspect ratio to be equal
        self.ax.set_aspect("equal", adjustable="box")

        # Hide axis
        plt.axis("off")

        # Display the plot
        plt.show()



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing models to be used in benchmarks."""

from qcml.models.quantum.circuit_centric import CircuitCentricClassifier
from qcml.models.quantum.convolutional_neural_network import (
    ConvolutionalNeuralNetwork,
)
from qcml.models.quantum.data_reuploading import (
    DataReuploadingClassifier,
    DataReuploadingClassifierNoScaling,
    DataReuploadingClassifierNoCost,
    DataReuploadingClassifierNoTrainableEmbedding,
    DataReuploadingClassifierSeparable,
)
from qcml.models.quantum.dressed_quantum_circuit import (
    DressedQuantumCircuitClassifier,
    DressedQuantumCircuitClassifierOnlyNN,
    DressedQuantumCircuitClassifierSeparable,
)

from qcml.models.quantum.iqp_kernel import IQPKernelClassifier
from qcml.models.quantum.iqp_variational import IQPVariationalClassifier
from qcml.models.quantum.projected_quantum_kernel import ProjectedQuantumKernel
from qcml.models.quantum.quantum_boltzmann_machine import (
    QuantumBoltzmannMachine,
    QuantumBoltzmannMachineSeparable,
)
from qcml.models.quantum.quantum_kitchen_sinks import QuantumKitchenSinks
from qcml.models.quantum.quantum_metric_learning import QuantumMetricLearner
from qcml.models.quantum.quanvolutional_neural_network import (
    QuanvolutionalNeuralNetwork,
)
from qcml.models.quantum.separable import (
    SeparableVariationalClassifier,
    SeparableKernelClassifier,
)
from qcml.models.quantum.tree_tensor import TreeTensorClassifier
from qcml.models.quantum.vanilla_qnn import VanillaQNN
from qcml.models.quantum.weinet import WeiNet

from qcml.models.classical.svc import SVCCustom as SVCCustom_Base
from qcml.models.classical.perceptron import Perceptron as PerceptronCustom_base
from qcml.models.classical.kernel_perceptron import (
    KernelPerceptron as KernelPerceptron_base,
)
from qcml.models.classical.kernel_mlp import KernelMLPClassifier

from qcml.models.quantum.ansatz_embedding_kernel import AnsatzEmbeddingKernel

from sklearn.svm import SVC as SVC_base
from sklearn.svm import LinearSVC as LinearSVC_base
from sklearn.linear_model import Perceptron as Perceptron_base
from sklearn.neural_network import MLPClassifier as MLP

__all__ = [
    "CircuitCentricClassifier",
    "ConvolutionalNeuralNetwork",
    "DataReuploadingClassifier",
    "DataReuploadingClassifierNoScaling",
    "DataReuploadingClassifierNoCost",
    "DataReuploadingClassifierNoTrainableEmbedding",
    "DataReuploadingClassifierSeparable",
    "DressedQuantumCircuitClassifier",
    "DressedQuantumCircuitClassifierOnlyNN",
    "DressedQuantumCircuitClassifierSeparable",
    "IQPKernelClassifier",
    "IQPVariationalClassifier",
    "ProjectedQuantumKernel",
    "QuantumBoltzmannMachine",
    "QuantumBoltzmannMachineSeparable",
    "QuantumKitchenSinks",
    "QuantumMetricLearner",
    "QuanvolutionalNeuralNetwork",
    "SeparableVariationalClassifier",
    "SeparableKernelClassifier",
    "TreeTensorClassifier",
    "VanillaQNN",
    "WeiNet",
    "MLPClassifier",
    "LinearSVC",
    "SVC",
    "SVCCustom",
    "Perceptron",
    "KernelPerceptron",
    "KernelMLPClassifier",
    "AnsatzEmbeddingKernel",
]


class MLPClassifier(MLP):
    def __init__(
        self,
        hidden_layer_sizes=(
            100,
            100,
        ),  # The ith element represents the number of neurons in the ith hidden layer
        activation="relu",  # Activation function for the hidden layer ('relu', 'logistic', 'tanh', 'identity')
        solver="adam",  # The solver for weight optimization ('lbfgs', 'sgd', 'adam')
        alpha=0.0001,  # L2 penalty (regularization term) parameter
        batch_size="auto",  # Size of minibatches for stochastic optimizers
        learning_rate="constant",  # Learning rate schedule for weight updates ('constant', 'invscaling', 'adaptive')
        learning_rate_init=0.001,  # Initial learning rate used
        power_t=0.5,  # The exponent for inverse scaling learning rate
        max_iter=3000,  # Maximum number of iterations
        shuffle=True,  # Whether to shuffle samples in each iteration
        random_state=None,  # Seed for the random number generator
        tol=1e-4,  # Tolerance for the optimization
        verbose=False,  # Whether to print progress messages to stdout
        warm_start=False,  # Whether to reuse the solution of the previous call to fit
        momentum=0.9,  # Momentum for gradient descent update
        nesterovs_momentum=True,  # Whether to use Nesterov’s momentum
        early_stopping=False,  # Whether to use early stopping to terminate training when validation score is not improving
        validation_fraction=0.1,  # The proportion of training data to set aside as validation set for early stopping
        beta_1=0.9,  # Exponential decay rate for estimates of first moment vector in Adam
        beta_2=0.999,  # Exponential decay rate for estimates of second moment vector in Adam
        epsilon=1e-8,  # Value for numerical stability in Adam
        n_iter_no_change=10,  # Number of iterations with no improvement to wait before early stopping
        max_fun=15000,  # Maximum number of function calls in the optimizer
    ):
        super().__init__(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            solver=solver,
            alpha=alpha,
            batch_size=batch_size,
            learning_rate=learning_rate,
            learning_rate_init=learning_rate_init,
            power_t=power_t,
            max_iter=max_iter,
            shuffle=shuffle,
            random_state=random_state,
            tol=tol,
            verbose=verbose,
            warm_start=warm_start,
            momentum=momentum,
            nesterovs_momentum=nesterovs_momentum,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            beta_1=beta_1,
            beta_2=beta_2,
            epsilon=epsilon,
            n_iter_no_change=n_iter_no_change,
            max_fun=max_fun,
        )


class MLPClassifierCustom(KernelMLPClassifier):
    def __init__(
        self,
        hidden_layer_sizes=(
            100,
            100,
        ),  # The ith element represents the number of neurons in the ith hidden layer
        kernel_layer=None,  # Dictionary for kernel layer parameters (None means no kernel layer)
        kernel_func=None,  # Kernel function (if using kernel layer)
        kernel_params=None,  # Parameters for the kernel function (if using kernel layer)
        learning_rate=0.001,  # Learning rate for the optimizer
        num_classes=10,  # Number of output classes
        max_iter=3000,  # Maximum number of iterations
        batch_size=32,  # Batch size for training
        one_hot_encode=True,  # Whether to one-hot encode the labels
    ):
        # Store the parameters as attributes
        self.hidden_layer_sizes = hidden_layer_sizes
        self.kernel_layer = kernel_layer
        self.kernel_func = kernel_func
        self.kernel_params = kernel_params
        self.learning_rate = learning_rate
        self.num_classes = num_classes
        self.max_iter = max_iter
        self.batch_size = batch_size
        self.one_hot_encode = one_hot_encode

        # Construct the layer configuration based on the provided parameters
        layer_config = []

        if kernel_layer is not None:
            layer_config.append(
                (
                    "kernel",
                    {
                        "n_centers": kernel_layer.get("n_centers", 10),
                        "kernel_func": kernel_func,
                        "kernel_params": kernel_params,
                    },
                )
            )

        # Add dense layers
        for size in hidden_layer_sizes:
            layer_config.append(("dense", {"features": size}))

        # Initialize the base class with the constructed configuration
        super().__init__(
            layer_config=layer_config,
            learning_rate=learning_rate,
            num_classes=num_classes,
            num_epochs=max_iter,
            batch_size=batch_size,
            one_hot_encode=one_hot_encode,
        )

    def get_params(self, deep=True):
        return {
            "hidden_layer_sizes": self.hidden_layer_sizes,
            "kernel_layer": self.kernel_layer,
            "kernel_func": self.kernel_func,
            "kernel_params": self.kernel_params,
            "learning_rate": self.learning_rate,
            "num_classes": self.num_classes,
            "max_iter": self.max_iter,
            "batch_size": self.batch_size,
            "one_hot_encode": self.one_hot_encode,
        }

    def set_params(self, **params):
        for key, value in params.items():
            setattr(self, key, value)
        return self


class LinearSVC(LinearSVC_base):
    def __init__(
        self,
        C=1.0,  # Regularization parameter. The strength of the regularization is inversely proportional to C
        loss="squared_hinge",  # Specifies the loss function, default is 'squared_hinge'
        tol=0.0001,  # Tolerance for stopping criterion
        fit_intercept=True,  # Whether to calculate the intercept for this model
        intercept_scaling=1,  # When fit_intercept=True, instance vector x becomes [x, intercept_scaling]
        class_weight=None,  # Weights associated with classes
        verbose=0,  # Enable verbose output
        random_state=None,  # Seed for the random number generator
        max_iter=2000,  # Hard limit on iterations within solver
    ):
        super().__init__(
            C=C,
            loss=loss,
            tol=tol,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            verbose=verbose,
            random_state=random_state,
            max_iter=max_iter,
        )


class SVC(SVC_base):
    def __init__(
        self,
        C=1.0,  # Regularization parameter. The strength of the regularization is inversely proportional to C
        kernel="precomputed",
        degree=3,  # Degree of the polynomial kernel function ('poly'). Ignored by all other kernels
        gamma="scale",  # Kernel coefficient for 'rbf', 'poly' and 'sigmoid'
        coef0=0.0,  # Independent term in kernel function. It is only significant in 'poly' and 'sigmoid'
        shrinking=True,  # Whether to use the shrinking heuristic
        probability=False,  # Whether to enable probability estimates
        tol=0.001,  # Tolerance for stopping criterion
        max_iter=2000,  # Hard limit on iterations within solver, or -1 for no limit
        random_state=None,  # Seed for the random number generator
    ):
        super().__init__(
            C=C,
            kernel=kernel,  # Specifies the kernel type to be used in the algorithm ('rbf' by default)
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            probability=probability,
            tol=tol,
            max_iter=max_iter,
            random_state=random_state,
        )


class Perceptron(Perceptron_base):
    def __init__(
        self,
        penalty=None,  # No penalty (no regularization)
        alpha=0.0001,  # Regularization strength
        l1_ratio=0.15,  # Ratio of L1 regularization (used only if penalty is 'elasticnet')
        fit_intercept=True,  # Whether to calculate the intercept for this model
        max_iter=2000,  # Maximum number of passes over the training data
        tol=0.001,  # The stopping criterion (tolerance)
        shuffle=True,  # Whether to shuffle the training data after each epoch
        verbose=0,  # The verbosity level
        eta0=1.0,  # Constant by which the updates are multiplied
        n_jobs=None,  # Number of CPU cores used during the cross-validation loop
        random_state=0,  # Seed for the random number generator
        early_stopping=False,  # Whether to use early stopping to terminate training
        validation_fraction=0.1,  # The proportion of training data to set aside as validation set
        n_iter_no_change=5,  # Number of iterations with no improvement to wait before early stopping
        class_weight=None,  # Weights associated with classes
        warm_start=False,  # Whether to reuse the solution of the previous call to fit
    ):
        super().__init__(
            penalty=penalty,
            alpha=alpha,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            max_iter=max_iter,
            tol=tol,
            shuffle=shuffle,
            verbose=verbose,
            eta0=eta0,
            n_jobs=n_jobs,
            random_state=random_state,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            class_weight=class_weight,
            warm_start=warm_start,
        )


class KernelPerceptron(KernelPerceptron_base):
    def __init__(
        self,
        kernel="precomputed",  # The kernel type to be used in the algorithm (default is "precomputed")
        n_iter=200,  # Number of passes over the training data
        random_state=42,  # Seed for the random number generator
        params=None,  # Additional parameters for the kernel function
        max_iter=1000,  # Maximum number of passes over the training data
    ):
        super().__init__(
            kernel=kernel,
            n_iter=n_iter,
            random_state=random_state,
            params=params,
            max_iter=max_iter,
        )


# Extend SVCCustom
class SVCCustom(SVCCustom_Base):
    def __init__(
        self,
        C=1.0,
        kernel="rbf",
        degree=3,
        gamma="scale",
        coef0=0.0,
        shrinking=True,
        probability=False,
        tol=0.001,
        max_iter=-1,
        random_state=None,
        **kernel_params,
    ):
        super().__init__(
            C=C,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            probability=probability,
            tol=tol,
            max_iter=max_iter,
            random_state=random_state,
            **kernel_params,
        )


class PerceptronCustom(PerceptronCustom_base):
    def __init__(self, n_iter=5, random_state=42, max_iter=1000):
        super().__init__(n_iter=n_iter, random_state=random_state, max_iter=max_iter)



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import pennylane as qml
import numpy as np
import jax
import jax.numpy as jnp
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.svm import SVC
from sklearn.preprocessing import MinMaxScaler
from qcml.utils.model import chunk_vmapped_fn
from qcml.models.classical import KernelPerceptron

jax.config.update("jax_enable_x64", True)


class IQPKernelClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        classifier=None,
        repeats=2,
        C=1.0,
        jit=False,
        random_state=42,
        scaling=1.0,
        max_vmap=250,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax-jit", "diff_method": None},
    ):
        r"""
        Kernel version of the classifier from https://arxiv.org/pdf/1804.11326v2.pdf.
        The kernel function is given by

        .. math::
            k(x,x')=\vert\langle 0 \vert U^\dagger(x')U(x)\vert 0 \rangle\vert^2

        where :math:`U(x)` is an IQP circuit implemented via Pennylane's `IQPEmbedding`.

        We precompute the kernel matrix from the data directly, and pass it to scikit-learn's support vector
        classifier SVC. This  allows us to benefit from JAX parallelisation when computing the kernel
        matrices.

        The model requires evaluating a number of circuits given by the square of the number of data
        samples, and is therefore only appropriate for relatively small datasets.

        Args:
            svm (sklearn.svm.SVC): scikit-learn SVM class object used to fit the model from the kernel matrix
            repeats (int): number of times the IQP structure is repeated in the embedding circuit.
            C (float): regularization parameter for SVC. Lower values imply stronger regularization.
            jit (bool): Whether to use just in time compilation.
            dev_type (str): string specifying the pennylane device type; e.g. 'default.qubit'.
            max_vmap (int or None): The maximum size of a chunk to vectorise over. Lower values use less memory.
                must divide batch_size.
            qnode_kwargs (str): the key word arguments passed to the circuit qnode.
            scaling (float): Factor by which to scale the input data.
            random_state (int): seed used for reproducibility.
        """
        # attributes that do not depend on data
        self.repeats = repeats
        self.C = C
        self.jit = jit
        self.max_vmap = max_vmap
        self.classifier = classifier
        self.dev_type = dev_type
        self.qnode_kwargs = qnode_kwargs
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None
        self.n_qubits_ = None
        self.scaler = None  # data scaler will be fitted on training data
        self.circuit = None

    def generate_key(self):
        return jax.random.PRNGKey(self.rng.integers(1000000))

    def construct_circuit(self):
        dev = qml.device(self.dev_type, wires=self.n_qubits_)

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(x):
            """
            circuit used to precomute the kernel matrix K(x_1,x_2).
            Args:
                x (np.array): vector of length 2*num_feature that is the concatenation of x_1 and x_2

            Returns:
                (float) the value of the kernel fucntion K(x_1,x_2)
            """
            qml.IQPEmbedding(
                x[: self.n_qubits_], wires=range(self.n_qubits_), n_repeats=self.repeats
            )
            qml.adjoint(
                qml.IQPEmbedding(
                    x[self.n_qubits_ :],
                    wires=range(self.n_qubits_),
                    n_repeats=self.repeats,
                )
            )
            return qml.probs()

        self.circuit = circuit

        if self.jit:
            circuit = jax.jit(circuit)
        return circuit

    def precompute_kernel(self, X1, X2):
        """
        compute the kernel matrix relative to data sets X1 and X2
        Args:
            X1 (np.array): first dataset of input vectors
            X2 (np.array): second dataset of input vectors
        Returns:
            kernel_matrix (np.array): matrix of size (len(X1),len(X2)) with elements K(x_1,x_2)
        """
        dim1 = len(X1)
        dim2 = len(X2)

        # concatenate all pairs of vectors
        Z = jnp.array(
            [np.concatenate((X1[i], X2[j])) for i in range(dim1) for j in range(dim2)]
        )

        circuit = self.construct_circuit()
        self.batched_circuit = chunk_vmapped_fn(
            jax.vmap(circuit, 0), start=0, max_vmap=self.max_vmap
        )
        kernel_values = self.batched_circuit(Z)[:, 0]

        # reshape the values into the kernel matrix
        kernel_matrix = np.reshape(kernel_values, (dim1, dim2))

        return kernel_matrix

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.

        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels that the classifier expects
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        self.n_qubits_ = n_features

        self.construct_circuit()

    def fit(self, X, y):
        """Fit the model to data X and labels y. Uses sklearn's SVM classifier

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """

        self.classifier.random_state = int(
            jax.random.randint(
                self.generate_key(), shape=(1,), minval=0, maxval=1000000
            )
        )

        self.initialize(X.shape[1], np.unique(y))

        self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
        self.scaler.fit(X)
        X = self.transform(X)

        self.params_ = {"x_train": X}
        kernel_matrix = self.precompute_kernel(X, X)

        start = time.time()
        # we are updating this value here, in case it was
        # changed after initialising the model
        if isinstance(self.classifier, SVC):
            self.classifier.C = self.C
        self.classifier.fit(kernel_matrix, y)
        end = time.time()
        self.training_time_ = end - start
        return self

    def predict(self, X):
        """Predict labels for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred (np.ndarray): Predicted labels of shape (n_samples,)
        """
        X = self.transform(X)
        kernel_matrix = self.precompute_kernel(X, self.params_["x_train"])
        return self.svm.predict(kernel_matrix)

    def predict_proba(self, X):
        """Predict label probabilities for data X.
        note that this may be inconsistent with predict; see the sklearn docummentation for details.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """

        if "x_train" not in self.params_:
            raise ValueError("Model cannot predict without fitting to data first.")

        X = self.transform(X)
        kernel_matrix = self.precompute_kernel(X, self.params_["x_train"])
        return self.svm.predict_proba(kernel_matrix)

    def transform(self, X, preprocess=True):
        """
        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        if preprocess:
            if self.scaler is None:
                # if the model is unfitted, initialise the scaler here
                self.scaler = MinMaxScaler(feature_range=(-np.pi / 2, np.pi / 2))
                self.scaler.fit(X)
            X = self.scaler.transform(X)

        return X * self.scaling
